# -*- coding: utf-8 -*-

from typing import List, Dict, Any
from pos_model import Order
from old_helper import remove_xml_namespace
from _ContextKeys import ContextKeys

from _NfceXmlPartBuilder import NfceXmlPartBuilder
from nfceutil import NfeSigner


class NfeBuilder(NfceXmlPartBuilder):
    _nfce_template = """<NFe xmlns="http://www.portalfiscal.inf.br/nfe">{0}</NFe>"""

    def __init__(self, part_builders, xml_signer, inf_nfe_compl_builder):
        # type: (List[NfceXmlPartBuilder], NfeSigner, NfceXmlPartBuilder) -> None
        self.part_builders = part_builders
        self.xml_signer = xml_signer
        self.inf_nfe_compl_builder = inf_nfe_compl_builder

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode

        inner_xml = ""
        for part_builder in self.part_builders:
            inner_xml += part_builder.build_xml(order, context)

        unsigned_nfe = self._nfce_template.format(inner_xml)

        signature = self.xml_signer.get_signature(unsigned_nfe)
        signature_xml = remove_xml_namespace(signature)
        context[ContextKeys.digest_value] = signature_xml.find(".//Reference/DigestValue").text

        inf_nfe_comp = self.inf_nfe_compl_builder.build_xml(order, context)

        inner_xml += inf_nfe_comp
        inner_xml += signature

        return self._nfce_template.format(inner_xml)



# mypy: ignore-errors


import importlib

from . import testing
from .. import assert_raises
from .. import config
from .. import engines
from .. import eq_
from .. import fixtures
from .. import is_not_none
from .. import is_true
from .. import ne_
from .. import provide_metadata
from ..assertions import expect_raises
from ..assertions import expect_raises_message
from ..config import requirements
from ..provision import set_default_schema_on_connection
from ..schema import Column
from ..schema import Table
from ... import bindparam
from ... import dialects
from ... import event
from ... import exc
from ... import Integer
from ... import literal_column
from ... import select
from ... import String
from ...sql.compiler import Compiled
from ...util import inspect_getfullargspec


class PingTest(fixtures.TestBase):
    __backend__ = True

    def test_do_ping(self):
        with testing.db.connect() as conn:
            is_true(
                testing.db.dialect.do_ping(conn.connection.dbapi_connection)
            )


class ArgSignatureTest(fixtures.TestBase):
    """test that all visit_XYZ() in :class:`_sql.Compiler` subclasses have
    ``**kw``, for #8988.

    This test uses runtime code inspection.   Does not need to be a
    ``__backend__`` test as it only needs to run once provided all target
    dialects have been imported.

    For third party dialects, the suite would be run with that third
    party as a "--dburi", which means its compiler classes will have been
    imported by the time this test runs.

    """

    def _all_subclasses():  # type: ignore  # noqa
        for d in dialects.__all__:
            if not d.startswith("_"):
                importlib.import_module("sqlalchemy.dialects.%s" % d)

        stack = [Compiled]

        while stack:
            cls = stack.pop(0)
            stack.extend(cls.__subclasses__())
            yield cls

    @testing.fixture(params=list(_all_subclasses()))
    def all_subclasses(self, request):
        yield request.param

    def test_all_visit_methods_accept_kw(self, all_subclasses):
        cls = all_subclasses

        for k in cls.__dict__:
            if k.startswith("visit_"):
                meth = getattr(cls, k)

                insp = inspect_getfullargspec(meth)
                is_not_none(
                    insp.varkw,
                    f"Compiler visit method {cls.__name__}.{k}() does "
                    "not accommodate for **kw in its argument signature",
                )


class ExceptionTest(fixtures.TablesTest):
    """Test basic exception wrapping.

    DBAPIs vary a lot in exception behavior so to actually anticipate
    specific exceptions from real round trips, we need to be conservative.

    """

    run_deletes = "each"

    __backend__ = True

    @classmethod
    def define_tables(cls, metadata):
        Table(
            "manual_pk",
            metadata,
            Column("id", Integer, primary_key=True, autoincrement=False),
            Column("data", String(50)),
        )

    @requirements.duplicate_key_raises_integrity_error
    def test_integrity_error(self):
        with config.db.connect() as conn:
            trans = conn.begin()
            conn.execute(
                self.tables.manual_pk.insert(), {"id": 1, "data": "d1"}
            )

            assert_raises(
                exc.IntegrityError,
                conn.execute,
                self.tables.manual_pk.insert(),
                {"id": 1, "data": "d1"},
            )

            trans.rollback()

    def test_exception_with_non_ascii(self):
        with config.db.connect() as conn:
            try:
                # try to create an error message that likely has non-ascii
                # characters in the DBAPI's message string.  unfortunately
                # there's no way to make this happen with some drivers like
                # mysqlclient, pymysql.  this at least does produce a non-
                # ascii error message for cx_oracle, psycopg2
                conn.execute(select(literal_column("méil")))
                assert False
            except exc.DBAPIError as err:
                err_str = str(err)

                assert str(err.orig) in str(err)

            assert isinstance(err_str, str)


class IsolationLevelTest(fixtures.TestBase):
    __backend__ = True

    __requires__ = ("isolation_level",)

    def _get_non_default_isolation_level(self):
        levels = requirements.get_isolation_levels(config)

        default = levels["default"]
        supported = levels["supported"]

        s = set(supported).difference(["AUTOCOMMIT", default])
        if s:
            return s.pop()
        else:
            config.skip_test("no non-default isolation level available")

    def test_default_isolation_level(self):
        eq_(
            config.db.dialect.default_isolation_level,
            requirements.get_isolation_levels(config)["default"],
        )

    def test_non_default_isolation_level(self):
        non_default = self._get_non_default_isolation_level()

        with config.db.connect() as conn:
            existing = conn.get_isolation_level()

            ne_(existing, non_default)

            conn.execution_options(isolation_level=non_default)

            eq_(conn.get_isolation_level(), non_default)

            conn.dialect.reset_isolation_level(
                conn.connection.dbapi_connection
            )

            eq_(conn.get_isolation_level(), existing)

    def test_all_levels(self):
        levels = requirements.get_isolation_levels(config)

        all_levels = levels["supported"]

        for level in set(all_levels).difference(["AUTOCOMMIT"]):
            with config.db.connect() as conn:
                conn.execution_options(isolation_level=level)

                eq_(conn.get_isolation_level(), level)

                trans = conn.begin()
                trans.rollback()

                eq_(conn.get_isolation_level(), level)

            with config.db.connect() as conn:
                eq_(
                    conn.get_isolation_level(),
                    levels["default"],
                )

    @testing.requires.get_isolation_level_values
    def test_invalid_level_execution_option(self, connection_no_trans):
        """test for the new get_isolation_level_values() method"""

        connection = connection_no_trans
        with expect_raises_message(
            exc.ArgumentError,
            "Invalid value '%s' for isolation_level. "
            "Valid isolation levels for '%s' are %s"
            % (
                "FOO",
                connection.dialect.name,
                ", ".join(
                    requirements.get_isolation_levels(config)["supported"]
                ),
            ),
        ):
            connection.execution_options(isolation_level="FOO")

    @testing.requires.get_isolation_level_values
    @testing.requires.dialect_level_isolation_level_param
    def test_invalid_level_engine_param(self, testing_engine):
        """test for the new get_isolation_level_values() method
        and support for the dialect-level 'isolation_level' parameter.

        """

        eng = testing_engine(options=dict(isolation_level="FOO"))
        with expect_raises_message(
            exc.ArgumentError,
            "Invalid value '%s' for isolation_level. "
            "Valid isolation levels for '%s' are %s"
            % (
                "FOO",
                eng.dialect.name,
                ", ".join(
                    requirements.get_isolation_levels(config)["supported"]
                ),
            ),
        ):
            eng.connect()

    @testing.requires.independent_readonly_connections
    def test_dialect_user_setting_is_restored(self, testing_engine):
        levels = requirements.get_isolation_levels(config)
        default = levels["default"]
        supported = (
            sorted(
                set(levels["supported"]).difference([default, "AUTOCOMMIT"])
            )
        )[0]

        e = testing_engine(options={"isolation_level": supported})

        with e.connect() as conn:
            eq_(conn.get_isolation_level(), supported)

        with e.connect() as conn:
            conn.execution_options(isolation_level=default)
            eq_(conn.get_isolation_level(), default)

        with e.connect() as conn:
            eq_(conn.get_isolation_level(), supported)


class AutocommitIsolationTest(fixtures.TablesTest):
    run_deletes = "each"

    __requires__ = ("autocommit",)

    __backend__ = True

    @classmethod
    def define_tables(cls, metadata):
        Table(
            "some_table",
            metadata,
            Column("id", Integer, primary_key=True, autoincrement=False),
            Column("data", String(50)),
            test_needs_acid=True,
        )

    def _test_conn_autocommits(self, conn, autocommit):
        trans = conn.begin()
        conn.execute(
            self.tables.some_table.insert(), {"id": 1, "data": "some data"}
        )
        trans.rollback()

        eq_(
            conn.scalar(select(self.tables.some_table.c.id)),
            1 if autocommit else None,
        )
        conn.rollback()

        with conn.begin():
            conn.execute(self.tables.some_table.delete())

    def test_autocommit_on(self, connection_no_trans):
        conn = connection_no_trans
        c2 = conn.execution_options(isolation_level="AUTOCOMMIT")
        self._test_conn_autocommits(c2, True)

        c2.dialect.reset_isolation_level(c2.connection.dbapi_connection)

        self._test_conn_autocommits(conn, False)

    def test_autocommit_off(self, connection_no_trans):
        conn = connection_no_trans
        self._test_conn_autocommits(conn, False)

    def test_turn_autocommit_off_via_default_iso_level(
        self, connection_no_trans
    ):
        conn = connection_no_trans
        conn = conn.execution_options(isolation_level="AUTOCOMMIT")
        self._test_conn_autocommits(conn, True)

        conn.execution_options(
            isolation_level=requirements.get_isolation_levels(config)[
                "default"
            ]
        )
        self._test_conn_autocommits(conn, False)

    @testing.requires.independent_readonly_connections
    @testing.variation("use_dialect_setting", [True, False])
    def test_dialect_autocommit_is_restored(
        self, testing_engine, use_dialect_setting
    ):
        """test #10147"""

        if use_dialect_setting:
            e = testing_engine(options={"isolation_level": "AUTOCOMMIT"})
        else:
            e = testing_engine().execution_options(
                isolation_level="AUTOCOMMIT"
            )

        levels = requirements.get_isolation_levels(config)

        default = levels["default"]

        with e.connect() as conn:
            self._test_conn_autocommits(conn, True)

        with e.connect() as conn:
            conn.execution_options(isolation_level=default)
            self._test_conn_autocommits(conn, False)

        with e.connect() as conn:
            self._test_conn_autocommits(conn, True)


class EscapingTest(fixtures.TestBase):
    @provide_metadata
    def test_percent_sign_round_trip(self):
        """test that the DBAPI accommodates for escaped / nonescaped
        percent signs in a way that matches the compiler

        """
        m = self.metadata
        t = Table("t", m, Column("data", String(50)))
        t.create(config.db)
        with config.db.begin() as conn:
            conn.execute(t.insert(), dict(data="some % value"))
            conn.execute(t.insert(), dict(data="some %% other value"))

            eq_(
                conn.scalar(
                    select(t.c.data).where(
                        t.c.data == literal_column("'some % value'")
                    )
                ),
                "some % value",
            )

            eq_(
                conn.scalar(
                    select(t.c.data).where(
                        t.c.data == literal_column("'some %% other value'")
                    )
                ),
                "some %% other value",
            )


class WeCanSetDefaultSchemaWEventsTest(fixtures.TestBase):
    __backend__ = True

    __requires__ = ("default_schema_name_switch",)

    def test_control_case(self):
        default_schema_name = config.db.dialect.default_schema_name

        eng = engines.testing_engine()
        with eng.connect():
            pass

        eq_(eng.dialect.default_schema_name, default_schema_name)

    def test_wont_work_wo_insert(self):
        default_schema_name = config.db.dialect.default_schema_name

        eng = engines.testing_engine()

        @event.listens_for(eng, "connect")
        def on_connect(dbapi_connection, connection_record):
            set_default_schema_on_connection(
                config, dbapi_connection, config.test_schema
            )

        with eng.connect() as conn:
            what_it_should_be = eng.dialect._get_default_schema_name(conn)
            eq_(what_it_should_be, config.test_schema)

        eq_(eng.dialect.default_schema_name, default_schema_name)

    def test_schema_change_on_connect(self):
        eng = engines.testing_engine()

        @event.listens_for(eng, "connect", insert=True)
        def on_connect(dbapi_connection, connection_record):
            set_default_schema_on_connection(
                config, dbapi_connection, config.test_schema
            )

        with eng.connect() as conn:
            what_it_should_be = eng.dialect._get_default_schema_name(conn)
            eq_(what_it_should_be, config.test_schema)

        eq_(eng.dialect.default_schema_name, config.test_schema)

    def test_schema_change_works_w_transactions(self):
        eng = engines.testing_engine()

        @event.listens_for(eng, "connect", insert=True)
        def on_connect(dbapi_connection, *arg):
            set_default_schema_on_connection(
                config, dbapi_connection, config.test_schema
            )

        with eng.connect() as conn:
            trans = conn.begin()
            what_it_should_be = eng.dialect._get_default_schema_name(conn)
            eq_(what_it_should_be, config.test_schema)
            trans.rollback()

            what_it_should_be = eng.dialect._get_default_schema_name(conn)
            eq_(what_it_should_be, config.test_schema)

        eq_(eng.dialect.default_schema_name, config.test_schema)


class FutureWeCanSetDefaultSchemaWEventsTest(
    fixtures.FutureEngineMixin, WeCanSetDefaultSchemaWEventsTest
):
    pass


class DifficultParametersTest(fixtures.TestBase):
    __backend__ = True

    tough_parameters = testing.combinations(
        ("boring",),
        ("per cent",),
        ("per % cent",),
        ("%percent",),
        ("par(ens)",),
        ("percent%(ens)yah",),
        ("col:ons",),
        ("_starts_with_underscore",),
        ("dot.s",),
        ("more :: %colons%",),
        ("_name",),
        ("___name",),
        ("[BracketsAndCase]",),
        ("42numbers",),
        ("percent%signs",),
        ("has spaces",),
        ("/slashes/",),
        ("more/slashes",),
        ("q?marks",),
        ("1param",),
        ("1col:on",),
        argnames="paramname",
    )

    @tough_parameters
    @config.requirements.unusual_column_name_characters
    def test_round_trip_same_named_column(
        self, paramname, connection, metadata
    ):
        name = paramname

        t = Table(
            "t",
            metadata,
            Column("id", Integer, primary_key=True),
            Column(name, String(50), nullable=False),
        )

        # table is created
        t.create(connection)

        # automatic param generated by insert
        connection.execute(t.insert().values({"id": 1, name: "some name"}))

        # automatic param generated by criteria, plus selecting the column
        stmt = select(t.c[name]).where(t.c[name] == "some name")

        eq_(connection.scalar(stmt), "some name")

        # use the name in a param explicitly
        stmt = select(t.c[name]).where(t.c[name] == bindparam(name))

        row = connection.execute(stmt, {name: "some name"}).first()

        # name works as the key from cursor.description
        eq_(row._mapping[name], "some name")

        # use expanding IN
        stmt = select(t.c[name]).where(
            t.c[name].in_(["some name", "some other_name"])
        )

        row = connection.execute(stmt).first()

    @testing.fixture
    def multirow_fixture(self, metadata, connection):
        mytable = Table(
            "mytable",
            metadata,
            Column("myid", Integer),
            Column("name", String(50)),
            Column("desc", String(50)),
        )

        mytable.create(connection)

        connection.execute(
            mytable.insert(),
            [
                {"myid": 1, "name": "a", "desc": "a_desc"},
                {"myid": 2, "name": "b", "desc": "b_desc"},
                {"myid": 3, "name": "c", "desc": "c_desc"},
                {"myid": 4, "name": "d", "desc": "d_desc"},
            ],
        )
        yield mytable

    @tough_parameters
    def test_standalone_bindparam_escape(
        self, paramname, connection, multirow_fixture
    ):
        tbl1 = multirow_fixture
        stmt = select(tbl1.c.myid).where(
            tbl1.c.name == bindparam(paramname, value="x")
        )
        res = connection.scalar(stmt, {paramname: "c"})
        eq_(res, 3)

    @tough_parameters
    def test_standalone_bindparam_escape_expanding(
        self, paramname, connection, multirow_fixture
    ):
        tbl1 = multirow_fixture
        stmt = (
            select(tbl1.c.myid)
            .where(tbl1.c.name.in_(bindparam(paramname, value=["a", "b"])))
            .order_by(tbl1.c.myid)
        )

        res = connection.scalars(stmt, {paramname: ["d", "a"]}).all()
        eq_(res, [1, 4])


class ReturningGuardsTest(fixtures.TablesTest):
    """test that the various 'returning' flags are set appropriately"""

    __backend__ = True

    @classmethod
    def define_tables(cls, metadata):
        Table(
            "t",
            metadata,
            Column("id", Integer, primary_key=True, autoincrement=False),
            Column("data", String(50)),
        )

    @testing.fixture
    def run_stmt(self, connection):
        t = self.tables.t

        def go(stmt, executemany, id_param_name, expect_success):
            stmt = stmt.returning(t.c.id)

            if executemany:
                if not expect_success:
                    # for RETURNING executemany(), we raise our own
                    # error as this is independent of general RETURNING
                    # support
                    with expect_raises_message(
                        exc.StatementError,
                        rf"Dialect {connection.dialect.name}\+"
                        f"{connection.dialect.driver} with "
                        f"current server capabilities does not support "
                        f".*RETURNING when executemany is used",
                    ):
                        result = connection.execute(
                            stmt,
                            [
                                {id_param_name: 1, "data": "d1"},
                                {id_param_name: 2, "data": "d2"},
                                {id_param_name: 3, "data": "d3"},
                            ],
                        )
                else:
                    result = connection.execute(
                        stmt,
                        [
                            {id_param_name: 1, "data": "d1"},
                            {id_param_name: 2, "data": "d2"},
                            {id_param_name: 3, "data": "d3"},
                        ],
                    )
                    eq_(result.all(), [(1,), (2,), (3,)])
            else:
                if not expect_success:
                    # for RETURNING execute(), we pass all the way to the DB
                    # and let it fail
                    with expect_raises(exc.DBAPIError):
                        connection.execute(
                            stmt, {id_param_name: 1, "data": "d1"}
                        )
                else:
                    result = connection.execute(
                        stmt, {id_param_name: 1, "data": "d1"}
                    )
                    eq_(result.all(), [(1,)])

        return go

    def test_insert_single(self, connection, run_stmt):
        t = self.tables.t

        stmt = t.insert()

        run_stmt(stmt, False, "id", connection.dialect.insert_returning)

    def test_insert_many(self, connection, run_stmt):
        t = self.tables.t

        stmt = t.insert()

        run_stmt(
            stmt, True, "id", connection.dialect.insert_executemany_returning
        )

    def test_update_single(self, connection, run_stmt):
        t = self.tables.t

        connection.execute(
            t.insert(),
            [
                {"id": 1, "data": "d1"},
                {"id": 2, "data": "d2"},
                {"id": 3, "data": "d3"},
            ],
        )

        stmt = t.update().where(t.c.id == bindparam("b_id"))

        run_stmt(stmt, False, "b_id", connection.dialect.update_returning)

    def test_update_many(self, connection, run_stmt):
        t = self.tables.t

        connection.execute(
            t.insert(),
            [
                {"id": 1, "data": "d1"},
                {"id": 2, "data": "d2"},
                {"id": 3, "data": "d3"},
            ],
        )

        stmt = t.update().where(t.c.id == bindparam("b_id"))

        run_stmt(
            stmt, True, "b_id", connection.dialect.update_executemany_returning
        )

    def test_delete_single(self, connection, run_stmt):
        t = self.tables.t

        connection.execute(
            t.insert(),
            [
                {"id": 1, "data": "d1"},
                {"id": 2, "data": "d2"},
                {"id": 3, "data": "d3"},
            ],
        )

        stmt = t.delete().where(t.c.id == bindparam("b_id"))

        run_stmt(stmt, False, "b_id", connection.dialect.delete_returning)

    def test_delete_many(self, connection, run_stmt):
        t = self.tables.t

        connection.execute(
            t.insert(),
            [
                {"id": 1, "data": "d1"},
                {"id": 2, "data": "d2"},
                {"id": 3, "data": "d3"},
            ],
        )

        stmt = t.delete().where(t.c.id == bindparam("b_id"))

        run_stmt(
            stmt, True, "b_id", connection.dialect.delete_executemany_returning
        )



from ._PaymentProcessor import PaymentProcessor

from .cappta import (
    CapptaCreditPaymentProcessor,
    CapptaDebitPaymentProcessor,
)
from .connectcar import ConectCarPaymentProcessor
from .fake import FakePaymentProcessor
from .mobile import MobilePaymentProcessor
from .sitef import (
    SitefDebitPaymentProcessor,
    SitefCreditPaymentProcessor,
    SitefDigitalWalletsPaymentProcessor,
    SitefGiftCardPaymentProcessor,
)






# Copyright (C) 2002-2006 Python Software Foundation
# Author: Barry Warsaw
# Contact: email-sig@python.org

"""Base class for MIME type messages that are not multipart."""
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

__all__ = ['MIMENonMultipart']

from future.backports.email import errors
from future.backports.email.mime.base import MIMEBase


class MIMENonMultipart(MIMEBase):
    """Base class for MIME multipart/* type messages."""

    def attach(self, payload):
        # The public API prohibits attaching multiple subparts to MIMEBase
        # derived subtypes since none of them are, by definition, of content
        # type multipart/*
        raise errors.MultipartConversionError(
            'Cannot attach additional subparts to non-multipart/*')



from ._ProductionCommand import ProductionCommand


class ChangePathCommand(ProductionCommand):
    def __init__(self, path, enabled):
        super(ChangePathCommand, self).__init__(None, None)
        self.path = path
        self.enabled = enabled

    def get_hash_extra_value(self):
        # type: () -> str
        return "ChangePathCommand{}{}".format(self.path, self.enabled)



"""
Compatibility layer with Python 3.8/3.9
"""
from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:  # pragma: no cover
    # Prevent circular imports on runtime.
    from . import Distribution, EntryPoint
else:
    Distribution = EntryPoint = Any


def normalized_name(dist: Distribution) -> Optional[str]:
    """
    Honor name normalization for distributions that don't provide ``_normalized_name``.
    """
    try:
        return dist._normalized_name
    except AttributeError:
        from . import Prepared  # -> delay to prevent circular imports.

        return Prepared.normalize(getattr(dist, "name", None) or dist.metadata['Name'])


def ep_matches(ep: EntryPoint, **params) -> bool:
    """
    Workaround for ``EntryPoint`` objects without the ``matches`` method.
    """
    try:
        return ep.matches(**params)
    except AttributeError:
        from . import EntryPoint  # -> delay to prevent circular imports.

        # Reconstruct the EntryPoint object to make sure it is compatible.
        return EntryPoint(ep.name, ep.value, ep.group).matches(**params)



from logging import Logger

from production.command import ProductionCommandProcessor
from production.manager import OrderChanger
from production.treebuilder import BoxConfiguration
from production.view import ProductionView
from typing import Dict

from production.box import ViewBox


class ViewBoxBuilder(object):
    def __init__(self, command_processor, order_changer, views, loggers):
        # type: (ProductionCommandProcessor, OrderChanger, Dict[str, ProductionView], Dict[str, Logger]) -> None
        self.command_processor = command_processor
        self.order_changer = order_changer
        self.views = views
        self.loggers = loggers

    def build(self, config):
        # type: (BoxConfiguration) -> ViewBox
        if "View" not in config.extra_config:
            raise Exception("View box without a configured View: {}".format(config.name))

        tag_from = "TOTALED" if "TagFrom" not in config.extra_config \
            else config.extra_config["TagFrom"]
        prod_state_from = "TOTALED" if "ProdStateFrom" not in config.extra_config \
            else config.extra_config["ProdStateFrom"]
        return ViewBox(config.name,
                       self.command_processor,
                       self.order_changer,
                       self.views[config.extra_config["View"]],
                       tag_from,
                       prod_state_from,
                       self.loggers[config.log_level])



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from threading import Lock
from typing import Callable


class Once:
    """Execute a function exactly once and block all callers until the function returns

    Same as golang's `sync.Once <https://pkg.go.dev/sync#Once>`_
    """

    def __init__(self) -> None:
        self._lock = Lock()
        self._done = False

    def do_once(self, func: Callable[[], None]) -> bool:
        """Execute ``func`` if it hasn't been executed or return.

        Will block until ``func`` has been called by one thread.

        Returns:
            Whether or not ``func`` was executed in this call
        """

        # fast path, try to avoid locking
        if self._done:
            return False

        with self._lock:
            if not self._done:
                func()
                self._done = True
                return True
        return False



import logging
from multiprocessing import Lock

from modelcommon.events import DialogEventType
from sysactions import close_asynch_dialog, show_messagebox

message_lock_dict = Lock()
message_lock = {}
wait_dlg_id = {}


class MessageDialogInteractor(object):
    def __init__(self, params, event_type=None):
        # type: (dict, str) -> None

        self.pos_id = str(params.get("pos_id", None))
        self.params = params
        self.event_type = event_type
        self.logger = logging.getLogger("PosActions")

    def wait_messages(self):
        # type: () -> None

        global wait_dlg_id

        try:
            with self._get_pos_lock(self.pos_id):
                message = self.params.get("message", None)
                if self.event_type == DialogEventType.SHOW_WAIT_SCREEN:
                    message_id = show_messagebox(self.pos_id, message, timeout=120000, asynch=True, buttons="")
                    wait_dlg_id[self.pos_id] = message_id
                if self.event_type == DialogEventType.HIDE_WAIT_SCREEN and wait_dlg_id[self.pos_id] is not None:
                    close_asynch_dialog(self.pos_id,  wait_dlg_id[self.pos_id])
                    del wait_dlg_id[self.pos_id]
        except Exception as _:
            self.logger.exception("Error wait messages")

    @staticmethod
    def _get_pos_lock(pos_id):
        # type: (str) -> None

        with message_lock_dict:
            if pos_id not in message_lock:
                message_lock[pos_id] = Lock()
            return message_lock[pos_id]



import ast
import uuid
from datetime import datetime
from sqlite3 import Connection
from typing import (
    List,
    Optional,
)
from uuid import UUID

import iso8601
from dbapitemplate import DbApiTemplate
from messagebus import Event
from model import (
    EventPublishDetail,
    AsyncTarget,
    PublicationHistory,
    PublicationAction,
)
from model.repository import EventRepository


def convert_optional_to_date(data):
    if data is None:
        return None
    return iso8601.parse_date(data).replace(tzinfo=None)


class SqliteEventRepository(EventRepository):
    def __init__(self, conn: Connection):
        self.conn = DbApiTemplate(conn)

    def save_event(self, event: EventPublishDetail):
        self.conn.execute(
            sql=self._save_event_query,
            params=(
                str(event.id),
                self._optional_datetime_formatter(event.created_at),
                event.external_code,
                event.event.subject,
                event.event.event_type,
                event.event.data,
                str(event.event.headers),
                self._optional_datetime_formatter(event.next_retry_at),
                event.retry_index,
                event.target,
                event.async_target_reply_subject,
                event.async_target_id_field,
                event.async_target_id,
                event.fifo,
            )
        )

    def get_by_id(self, id: UUID):
        data = self.conn.execute(
            sql=self._get_by_id_query,
            params=(str(id), ),
            mapper=self._event_definition_mapper,
        )
        if len(data) == 0:
            return None

        history = self.conn.execute(
            sql=self._get_history_by_definition_id,
            params=(str(id), ),
            mapper=self._publication_history_mapper,
        )
        data[0].publication_history = history

        return data[0]

    def get_pending(self):

        return self.conn.execute(
            sql=self._get_pending_events_query,
            mapper=self._event_definition_mapper,
        )

    def get_abandoned(self) -> List[EventPublishDetail]:

        return self.conn.execute(
            sql=self._get_abandoned_events_query,
            mapper=self._event_definition_mapper,
        )

    def save(self, event_definition: EventPublishDetail) -> None:
        self.conn.execute(
            sql=self._update_event_definition_query,
            params=(
                self._optional_datetime_formatter(event_definition.acked_at),
                event_definition.retry_index,
                self._optional_datetime_formatter(event_definition.next_retry_at),
                self._optional_datetime_formatter(event_definition.abandoned_at),
                str(event_definition.id),
            )
        )
        for history in event_definition.publication_history:
            if hasattr(history, "_saved"):
                continue

            history._saved = True
            self.conn.execute(
                sql=self._create_publication_history,
                params=(
                    str(event_definition.id),
                    history.created_at,
                    history.action.name,
                )
            )

    def get_by_external_code(self, external_code: str):
        return self.conn.first(
            sql=self._get_by_external_code_query,
            mapper=self._event_definition_mapper,
            params=(external_code, )
        )

    def cleanup_abandoned_events(
        self,
        min_abandoned_date: datetime,
    ) -> None:
        ids = self.conn.execute(
            sql=self._get_cleanup_abandoned_events_ids,
            mapper=self._id_mapper,
            params=(min_abandoned_date.isoformat(),),
        )

        if len(ids) > 0:
            self._clean_events_by_ids(ids=ids)

    def cleanup_acked_events(
        self,
        min_acked_date: datetime,
    ) -> None:
        ids = self.conn.execute(
            sql=self._get_cleanup_acked_event_ids,
            mapper=self._id_mapper,
            params=(min_acked_date.isoformat(),),
        )

        if len(ids) > 0:
            self._clean_events_by_ids(ids=ids)

    def _clean_events_by_ids(self, ids: List[str]):
        delete_ids = f'"{",".join(ids)}"'

        query = self._cleanup_publication_history.format(delete_ids)
        self.conn.execute(
            sql=query,
        )
        query = self._cleanup_pending_events.format(delete_ids)
        self.conn.execute(
            sql=query,
        )

    @staticmethod
    def _id_mapper(row):
        return str(row["Id"])

    @staticmethod
    def _optional_datetime_formatter(v: Optional[datetime]) -> Optional[str]:
        if v is None:
            return v

        return v.isoformat()

    @staticmethod
    def _event_definition_mapper(row):
        async_target = None
        if row["AsyncReplySubject"] is not None:
            async_target = AsyncTarget(
                reply_subject=row["AsyncReplySubject"],
                id_field=row["AsyncIdField"]
            )
        return EventPublishDetail(
            id=uuid.UUID(row["Id"]),
            created_at=iso8601.parse_date(row["CreatedAt"]).replace(tzinfo=None),
            external_code=row["ExternalCode"],
            event=Event(
                subject=row["Subject"],
                event_type=row["EventType"],
                data=row["Data"],
                headers=ast.literal_eval(row["Headers"]),
            ),
            target=row["Target"],
            async_target=async_target,
            fifo=bool(row["Fifo"]),
            retry_index=row["RetryIndex"],
            acked_at=convert_optional_to_date(row["AckedAt"]),
            next_retry_at=convert_optional_to_date(row["NextRetryAt"]),
            abandoned_at=convert_optional_to_date(row["AbandonedAt"]),
        )

    @staticmethod
    def _publication_history_mapper(row):
        return PublicationHistory(
            created_at=iso8601.parse_date(row["CreatedAt"]).replace(tzinfo=None),
            action=PublicationAction(row["Action"])
        )

    _event_definition_columns = """
        Id,
        CreatedAt,
        ExternalCode,
        Subject,
        EventType,
        Data,
        Headers,
        AckedAt,
        RetryIndex,
        NextRetryAt,
        AbandonedAt,
        Target,
        AsyncReplySubject,
        AsyncIdField,
        Fifo
    """

    _save_event_query = """
        INSERT INTO PendingEvent (
            Id,
            CreatedAt,
            ExternalCode,
            Subject,
            EventType,
            Data,
            Headers,
            NextRetryAt,
            RetryIndex,
            Target,
            AsyncReplySubject,
            AsyncIdField,
            AsyncId,
            Fifo
        )
        VALUES
        (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
    """

    _get_by_id_query = f"""
        SELECT {_event_definition_columns}
        FROM
        PendingEvent
        WHERE Id = ?
    """

    _get_by_external_code_query = f"""
            SELECT {_event_definition_columns}
            FROM
            PendingEvent
            WHERE ExternalCode = ?
        """

    _get_pending_events_query = f"""
       SELECT {_event_definition_columns}
       FROM
       PendingEvent
       WHERE AckedAt IS NULL and AbandonedAt is NULL
       ORDER BY CreatedAt
       LIMIT 100
    """

    _get_abandoned_events_query = f"""
           SELECT {_event_definition_columns}
           FROM
           PendingEvent
           WHERE AbandonedAt IS NOT NULL
           ORDER BY CreatedAt DESC
           LIMIT 10
        """

    _update_event_definition_query = """
        UPDATE
        PendingEvent
        set AckedAt = ?,
            RetryIndex = ?,
            NextRetryAt = ?,
            AbandonedAt = ?
        WHERE Id = ?
    """

    _create_publication_history = """
        INSERT INTO PublicationHistory (
            DefinitionId,
            CreatedAt,
            Action
        )
        VALUES
        (?, ?, ?)
    """

    _get_history_by_definition_id = """
        SELECT CreatedAt, Action
        FROM PublicationHistory
        WHERE DefinitionId = ?
    """

    _get_cleanup_abandoned_events_ids = """
        SELECT Id FROM
            PendingEvent
        WHERE AbandonedAt IS NOT NULL AND AbandonedAt < ?
    """

    _get_cleanup_acked_event_ids = """
        SELECT Id FROM
            PendingEvent
        WHERE AckedAt IS NOT NULL AND AckedAt < ?
    """

    _cleanup_publication_history = """
        DELETE FROM
            PublicationHistory
        WHERE DefinitionId IN ({})
    """

    _cleanup_pending_events = """
        DELETE FROM
            PendingEvent
        WHERE Id IN ({})
    """



from create_table_order import create_table_order
from store_table_service import store_table_service
from do_split_items import do_split_items
from do_split_items_equally import do_split_items_equally



from application.domain import TableService
from model.customexception import InvalidTipError


class DoChangeTipInteractor(object):
    def __init__(self, table_service):
        # type: (TableService) -> None
        self.table_service = table_service

    def execute(self, pos_id, table_id, percentage, amount):
        # type: (str, str, str, str) -> str

        if (percentage != "" and percentage != "0") and (amount != "" and amount != "0"):
            raise InvalidTipError(code=InvalidTipError.ErrorCode.CANNOT_USE_TIP_BY_AMOUNT_AND_PERCENTAGE)
        elif percentage == "" and amount == "":
            raise InvalidTipError(code=InvalidTipError.ErrorCode.TIP_AMOUNT_MUST_BE_VALID)

        if percentage != "":
            return self.table_service.change_service_tip(pos_id, table_id, percentage)
        if amount != "":
            return self.table_service.change_service_tip(pos_id, table_id, '', amount)



# -*- coding: utf-8 -*-

import json

from sysactions import (
    action,
    show_messagebox,
    close_asynch_dialog,
)

from typing import Callable
from utilfunctions import compare_objects

from ..customer import Customer
from ..customer.customer_model import AddCustomerRequest

dialog_id = None


def insert_or_update_customer_with_confirmation(
    pos_id,                # type: str
    customer_json,         # type: dict
):
    # type: (...) -> bool

    def confirmation_callback(callback_action):
        if callback_action == "INSERT":
            message = "$CUSTOMER_WILL_BE_INSERTED"
        else:
            message = "$CUSTOMER_WILL_BE_UPDATED"

        confirmation = show_messagebox(
            posid=pos_id,
            message=message,
            timeout=720000,
            buttons="$OK|$CANCEL",
        )
        return confirmation != 1

    return insert_or_update_customer(
        customer_json=customer_json,
        confirmation_callback=confirmation_callback,
    )


def insert_or_update_customer(
    customer_json,                  # type: dict
    confirmation_callback=None,     # type: Callable[[str], bool]
):
    # type: (...) -> bool

    customer_request = _create_customer_request(customer_json)

    customer = Customer().get_customer(phone=customer_request.phone)
    if not customer:
        if confirmation_callback is not None and not confirmation_callback("INSERT"):
            return False
        Customer().add_customer(add_customer_request=customer_request)
    else:
        same_customer = compare_objects(
            obj1=customer_request,
            obj2=_create_customer_request_to_compare(customer=json.loads(customer)),
        )
        if not same_customer and confirmation_callback is not None and not confirmation_callback("UPDATE"):
            return False
        Customer().update_customer(add_customer_request=customer_request)

    return True


def _create_customer_request(address_data_load):
    add_customer_request = AddCustomerRequest(
        address_data_load["CUSTOMER_PHONE"],
        address_data_load["CUSTOMER_DOC"],
        address_data_load["CUSTOMER_NAME"],
        address_data_load["POSTAL_CODE"],
        address_data_load["STATE"],
        address_data_load["CITY"],
        address_data_load["STREET_NAME"],
        address_data_load["NEIGHBORHOOD"],
        address_data_load["STREET_NUMBER"],
        address_data_load["COMPLEMENT"],
        address_data_load["ADDRESS_REFERENCE"]
    )

    return add_customer_request


def _create_customer_request_to_compare(customer):
    add_customer_request = AddCustomerRequest(
        customer["phone"],
        customer["document"],
        customer["name"],
        customer["address"]["zip_code"],
        customer["address"]["state"],
        customer["address"]["city"],
        customer["address"]["street"],
        customer["address"]["neighborhood"],
        customer["address"]["number"],
        customer["address"]["complement"],
        customer["address"]["reference_point"]
    )
    return add_customer_request


@action
def zip_code_search(pos_id):
    global dialog_id
    dialog_id = show_messagebox(pos_id, "$SEARCHING_ZIP_CODE", buttons="", asynch=True, timeout=50000)


@action
def zip_code_search_finished(pos_id, error="false"):
    close_asynch_dialog(pos_id, dialog_id)
    if error == "true":
        show_messagebox(pos_id, "$ERROR_SEARCHING_ZIP_CODE", buttons="OK")



# -*- coding: utf-8 -*-
from xml.etree import cElementTree as eTree
from typing import List

from sysactions import get_podtype, OrderTaker
from pos_model import POSTypeMapper


def set_pos_type(pos_ot, model, sale_type=None, pod_type=None):
    # type: (OrderTaker, eTree.XML, str, str) -> None

    if not pod_type:
        pod_type = get_podtype(model)

    pos_type = POSTypeMapper.get(sale_type, pod_type)
    pos_ot.setOrderCustomProperty("POS_TYPE", pos_type)


def get_pos_types():
    # type: () -> List[str]

    return POSTypeMapper.get_types()



"""
Fixer that adds ``from builtins import object`` if there is a line
like this:
    class Foo(object):
"""

from lib2to3 import fixer_base

from libfuturize.fixer_util import touch_import_top


class FixObject(fixer_base.BaseFix):

    PATTERN = u"classdef< 'class' NAME '(' name='object' ')' colon=':' any >"

    def transform(self, node, results):
        touch_import_top(u'builtins', 'object', node)



from typing import Optional, Any, Dict
from ._DataType import DataType


class Message(object):
    def __init__(
        self,                           # type: Message
        token,                          # type: int
        data_type=DataType.empty,       # type: int
        data=None,                      # type: bytes
        timeout=-1,                     # type: int
        headers=None,                   # type: Dict[str, str]
        ttl=-1,                         # type: int
    ):
        # type: (...) -> None
        self.token = token
        self.data_type = data_type
        self.data = data
        self.timeout = timeout
        self.headers = headers if headers is not None else {}
        self.ttl = ttl
        self.imp_message = None  # type: Any

    def __eq__(self, o):
        # type: (object) -> bool
        if not isinstance(o, Message):
            return False

        return (
            self.token == o.token
            and self.data_type == o.data_type
            and self.data == o.data
            and self.timeout == o.timeout
            and self.headers == o.headers
            and self.ttl == o.ttl
        )

    def __ne__(self, o):
        # type: (object) -> bool
        return not self.__eq__(o)

    def __str__(self):
        return str({
            "token": self.token,
            "data_type": self.data_type,
            "timeout": self.timeout,
            "headers": self.headers,
            "ttl": self.ttl,
        })



from typing import Any

from messagebus import (
    AckMessage,
    Event,
    Message,
    MessageBus,
    NakMessage,
)
from messageprocessorutil import UuidMessageProcessor

from services.interactor import BOHChangeNotificationConfigInteractor


class BOHChangeNotificationConfigProcessor(UuidMessageProcessor):
    def __init__(
        self,
        boh_change_notification_config_interactor,  # type: BOHChangeNotificationConfigInteractor
    ):
        # type: (...) -> None

        super(BOHChangeNotificationConfigProcessor, self).__init__()
        self.boh_change_notification_config_interactor = boh_change_notification_config_interactor

    def parse_data(
        self,
        data,           # type: str
        event_type,     # type: str
    ):
        # type: (...) -> None

        return None

    def call_business(
        self,
        request,    # type: None
    ):
        # type: (...) -> str

        return self.boh_change_notification_config_interactor.process_request()

    def format_response(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Event
        request,        # type: None
        result,         # type: None
    ):
        # type: (...) -> None

        return self._do_not_reply_to_asynchronous_event()

    def format_exception(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Event
        input_obj,      # type: None
        exception,      # type: Exception
    ):
        # type: (...) -> None

        return self._do_not_reply_to_asynchronous_event()

    def format_parse_exception(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Event
        data,           # type: Any
        exception,      # type: Exception
    ):
        # type: (...) -> None

        return self._do_not_reply_to_asynchronous_event()

    @staticmethod
    def _do_not_reply_to_asynchronous_event():
        return None



"""
Adds this import line:

    from builtins import XYZ

for each of the functions XYZ that is used in the module.
"""

from __future__ import unicode_literals

from lib2to3 import fixer_base
from lib2to3.pygram import python_symbols as syms
from lib2to3.fixer_util import Name, Call, in_special_context

from libfuturize.fixer_util import touch_import_top

# All builtins are:
#     from future.builtins.iterators import (filter, map, zip)
#     from future.builtins.misc import (ascii, chr, hex, input, isinstance, oct, open, round, super)
#     from future.types import (bytes, dict, int, range, str)
# We don't need isinstance any more.

replaced_builtins = '''filter map zip
                       ascii chr hex input next oct open round super
                       bytes dict int range str'''.split()

expression = '|'.join(["name='{0}'".format(name) for name in replaced_builtins])


class FixFutureBuiltins(fixer_base.BaseFix):
    BM_compatible = True
    run_order = 9

    # Currently we only match uses as a function. This doesn't match e.g.:
    #     if isinstance(s, str):
    #         ...
    PATTERN = """
              power<
                 ({0}) trailer< '(' args=[any] ')' >
              rest=any* >
              """.format(expression)

    def transform(self, node, results):
        name = results["name"]
        touch_import_top(u'builtins', name.value, node)
        # name.replace(Name(u"input", prefix=name.prefix))




# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import ceil, log2


class Buckets:

    # No method of this class is protected by locks because instances of this
    # class are only used in methods that are protected by locks themselves.

    def __init__(self):
        self._counts = [0]

        # The term index refers to the number of the exponential histogram bucket
        # used to determine its boundaries. The lower boundary of a bucket is
        # determined by base ** index and the upper boundary of a bucket is
        # determined by base ** (index + 1). index values are signedto account
        # for values less than or equal to 1.

        # self._index_* will all have values equal to a certain index that is
        # determined by the corresponding mapping _map_to_index function and
        # the value of the index depends on the value passed to _map_to_index.

        # Index of the 0th position in self._counts: self._counts[0] is the
        # count in the bucket with index self.__index_base.
        self.__index_base = 0

        # self.__index_start is the smallest index value represented in
        # self._counts.
        self.__index_start = 0

        # self.__index_start is the largest index value represented in
        # self._counts.
        self.__index_end = 0

    @property
    def index_start(self) -> int:
        return self.__index_start

    @index_start.setter
    def index_start(self, value: int) -> None:
        self.__index_start = value

    @property
    def index_end(self) -> int:
        return self.__index_end

    @index_end.setter
    def index_end(self, value: int) -> None:
        self.__index_end = value

    @property
    def index_base(self) -> int:
        return self.__index_base

    @index_base.setter
    def index_base(self, value: int) -> None:
        self.__index_base = value

    @property
    def counts(self):
        return self._counts

    def grow(self, needed: int, max_size: int) -> None:

        size = len(self._counts)
        bias = self.__index_base - self.__index_start
        old_positive_limit = size - bias

        # 2 ** ceil(log2(needed)) finds the smallest power of two that is larger
        # or equal than needed:
        # 2 ** ceil(log2(1)) == 1
        # 2 ** ceil(log2(2)) == 2
        # 2 ** ceil(log2(3)) == 4
        # 2 ** ceil(log2(4)) == 4
        # 2 ** ceil(log2(5)) == 8
        # 2 ** ceil(log2(6)) == 8
        # 2 ** ceil(log2(7)) == 8
        # 2 ** ceil(log2(8)) == 8
        new_size = min(2 ** ceil(log2(needed)), max_size)

        new_positive_limit = new_size - bias

        tmp = [0] * new_size
        tmp[new_positive_limit:] = self._counts[old_positive_limit:]
        tmp[0:old_positive_limit] = self._counts[0:old_positive_limit]
        self._counts = tmp

    @property
    def offset(self) -> int:
        return self.__index_start

    def __len__(self) -> int:
        if len(self._counts) == 0:
            return 0

        if self.__index_end == self.__index_start and self[0] == 0:
            return 0

        return self.__index_end - self.__index_start + 1

    def __getitem__(self, key: int) -> int:
        bias = self.__index_base - self.__index_start

        if key < bias:
            key += len(self._counts)

        key -= bias

        return self._counts[key]

    def downscale(self, amount: int) -> None:
        """
        Rotates, then collapses 2 ** amount to 1 buckets.
        """

        bias = self.__index_base - self.__index_start

        if bias != 0:

            self.__index_base = self.__index_start

            # [0, 1, 2, 3, 4] Original backing array

            self._counts = self._counts[::-1]
            # [4, 3, 2, 1, 0]

            self._counts = (
                self._counts[:bias][::-1] + self._counts[bias:][::-1]
            )
            # [3, 4, 0, 1, 2] This is a rotation of the backing array.

        size = 1 + self.__index_end - self.__index_start
        each = 1 << amount
        inpos = 0
        outpos = 0

        pos = self.__index_start

        while pos <= self.__index_end:
            mod = pos % each
            if mod < 0:
                mod += each

            index = mod

            while index < each and inpos < size:

                if outpos != inpos:
                    self._counts[outpos] += self._counts[inpos]
                    self._counts[inpos] = 0

                inpos += 1
                pos += 1
                index += 1

            outpos += 1

        self.__index_start >>= amount
        self.__index_end >>= amount
        self.__index_base = self.__index_start

    def increment_bucket(self, bucket_index: int, increment: int = 1) -> None:
        self._counts[bucket_index] += increment



from interactor import BaseInteractor
from model.log import LogEntry


class EventInteractor(BaseInteractor):
    def execute(self, log_entry):
        # type: (LogEntry) -> None

        if log_entry is None:
            return

        if log_entry.event_type == "HVCOMPSTATE":
            return

        if not log_entry.business_period:
            return

        event_to_dispatch = self.log_repository.insert_entry(log_entry)
        self.event_dispatcher.dispatch_event(event_to_dispatch)



import json

from typing import Optional

import bustoken
import sysactions

from msgbus import TK_SYS_ACK

from model.customexception import RemoteOrderStoreStatusError
from .. import logger, mb_context

TK_REMOTE_ORDER_GET_STORE_STATUS = bustoken.create_token(bustoken.MSGPRT_LOW, "37", "214")


@sysactions.action
def get_remote_order_status(
    _=None,             # type: Optional[str]
    default=False,      # type: Optional[bool]
):
    # type: (...) -> str

    default_response = json.dumps(
        dict(
            isOnline=False,
            isOpened=False,
            lastExternalContact=None,
            closedSince=None
        )
    )

    if default:
        return default_response

    try:
        msg = mb_context.MB_EasySendMessage("RemoteOrder", TK_REMOTE_ORDER_GET_STORE_STATUS, timeout=5 * 1000000)
        if msg.token == TK_SYS_ACK:
            return msg.data
        raise RemoteOrderStoreStatusError("Could not retrieve remote order status")
    except Exception as e:
        logger.error(e)
        return default_response



from xml.etree import cElementTree as eTree
from threading import Thread, Condition

from msgbus import TK_USERCTRL_GETINFO
from messagebus import Message

from production.model import User


class UsersInfoCache(Thread):
    def __init__(self, message_bus, logger):
        super(UsersInfoCache, self).__init__()

        self.message_bus = message_bus
        self.logger = logger

        self.users_information = {}

        self.running = True
        self.sleep_condition = Condition()
        self.sleep_time = 300

    def stop(self):
        self.running = False
        with self.sleep_condition:
            self.sleep_condition.notify_all()

    def sleep(self, time):
        with self.sleep_condition:
            self.sleep_condition.wait(time)

    def run(self):
        while self.running:
            self.renew_cache()
            self.sleep(self.sleep_time)

    def renew_cache(self):
        try:
            users = {}
            message = Message(TK_USERCTRL_GETINFO, timeout=30000000)
            response = self.message_bus.send_message("BOSS", message)
            for user in eTree.XML(response.data).findall("user"):
                user_id = user.get("UserId")
                user_name = user.get("LongName")
                user_level = user.get("Level")
                users[user_id] = User(user_id, user_name, user_level)

            self.users_information = users
        except Exception as _:
            self.users_information = {}

    def get_users_info(self):
        if not self.users_information:
            self.renew_cache()

        return self.users_information



import json
import os
import shutil
import time
from datetime import datetime
from threading import Lock
from typing import (
    Dict,
)
import xml.etree.ElementTree as eTree

from application.model import UpdateType
from application.model.configuration import Configurations
from application.model.customexception import (
    DisabledFunctionality,
    ErrorCreatingBackup,
    ErrorDeletingContent,
    ErrorRecoveringFolder,
    ErrorExtractingLoaders,
    ErrorDownloadingLoaders,
    RollbackError,
    NotInsideUpdateWindow,
    ErrorValidatingOpenedUsers,
    ErrorUpdatingUserRoles,
)
from application.repository.api import LoaderUpdaterAPIRepository
from application.repository.pos import (
    LoaderUpdaterPOSRepository,
    VersionsPOSRepository,
)
from application.service import (
    UpdaterService,
    StaticConfigurationService,
)
from application.util import (
    LoggerUtil,
    ValidationUtil,
    settings,
)
from systools import sys_parsefilepath

GENESIS_PATH = "../../genesis"

logger = LoggerUtil.get_logger_name(__name__)


class LoaderUpdaterService(object):
    def __init__(self, configs, working_on_update, api_repository, pos_repository, version_repository, updater_service):
        # type: (Configurations, Lock, LoaderUpdaterAPIRepository, LoaderUpdaterPOSRepository, VersionsPOSRepository, UpdaterService) -> None # noqa

        self.update_type = UpdateType.loader
        self.update_name = self.update_type.name
        self.enabled = configs.updaters[self.update_name].enabled
        self.local_configs = configs.updaters[self.update_name]
        self.store_id = configs.store_id

        self.working_on_update = working_on_update
        self.api_repository = api_repository
        self.pos_repository = pos_repository
        self.version_repository = version_repository
        self.updater_service = updater_service

        data_path = sys_parsefilepath("{HVDATADIR}")
        work_dir = sys_parsefilepath(configs.backup_directory)
        self.data_relative_path = "data"
        self.server_relative_path = "data/server"
        self.databases_relative_path = "data/server/databases"
        self.data_path = os.path.join(data_path, GENESIS_PATH, self.data_relative_path)
        self.server_path = os.path.join(self.data_path, "server")
        self.databases_path = os.path.join(data_path, GENESIS_PATH, self.databases_relative_path)
        default_work_path = os.path.join(work_dir, self.update_type.name)
        self.downloads_path = os.path.join(default_work_path, "downloads")
        self.backups_path = os.path.join(default_work_path, "backups")
        self.work_path = os.path.join(default_work_path, "work_path")
        self.payments_json = os.path.join(self.server_path, "bundles/payments.json")
        self.genesis_operator_levels_json = os.path.join(self.server_path, "bundles/operatorLevels.json")
        self.htdocs_path = os.path.join(data_path, 'htdocs')
        self.htdocs_kui_path = os.path.join(self.htdocs_path, 'kui')
        self.sui_static_path = os.path.join(data_path, 'htdocs/sui/static')
        self.loaders_zip = "loaders.zip"

        if not os.path.exists(self.backups_path):
            os.makedirs(self.backups_path)
        if not os.path.exists(self.downloads_path):
            os.makedirs(self.downloads_path)

    def update_loaders(
        self,                       # type: LoaderUpdaterService
        restart_hv=False,           # type: bool
        retry_on_error=False,       # type: bool
    ):
        # type: (...) -> None

        settings.set_update_control(self.update_name, None)
        if not self.enabled:
            settings.set_update_control(self.update_name, "$UPDATE_DISABLED")
            raise DisabledFunctionality()

        rollback_necessary = False
        backup_package_zip = ''

        try:
            logger.info("Waiting Updating Loaders")
            with self.working_on_update:
                logger.info("Updating loaders")
                settings.set_update_control(self.update_name, "$UPDATING_LOADERS")
                settings.set_update_control("updating", True)
                last_loader_update = self.pos_repository.get_last_loader_update()
                external_update_id = last_loader_update.server_update_identifier
                last_loader_update_id = int(last_loader_update.update_id)

                self._delete_directories()
                changed_pos_ctrl_type = self.download_and_validate_loaders()

                settings.set_update_control(self.update_name, "$LOADERS_BACKUP")
                backup_package_zip = self._create_backup()
                rollback_necessary = True

                settings.set_update_control(self.update_name, "$APPLYING_NEW_LOADERS")
                current_update_id = str(last_loader_update_id + 1)
                self.pos_repository.insert_new_loader_update(
                    update_id=current_update_id,
                    update_name="Loaders",
                    external_update_id=external_update_id,
                    obtained_date=last_loader_update.obtained_date,
                )

                self.pos_repository.delete_update_line(
                    external_update_id=external_update_id,
                    update_id=last_loader_update_id,
                )

                self._extract_new_loaders(self.data_path)
                self._recover_folders()

                settings.set_update_control(self.update_name, "$UPDATING_USER_FUNCTIONS")
                self.update_operator_levels()

                self.updater_service.update_databases_with_json_config(
                    update_name=self.update_name,
                    settings=settings,
                )

                self.pos_repository.mark_loader_update_as_applied()

                StaticConfigurationService.update_static_loaders()
                StaticConfigurationService.run_databases_script()
                settings.set_update_control(self.update_name, "$SUCCESS_UPDATE")

                if changed_pos_ctrl_type is not None:
                    self._update_pos_ctrl_type(changed_pos_ctrl_type)

                self._update_ui_customizations()

                rollback_necessary = False
                if external_update_id:
                    self.send_result_message(
                        current_update_id=current_update_id,
                        external_update_id=external_update_id,
                    )
                if restart_hv:
                    self._delete_directories()
                    logger.info("Starting Hypervisor to apply new Loaders")
                    self.pos_repository.common_pos_repository.restart_hv(
                        event_type="POS_UPDATER_LOADER",
                    )
                    time.sleep(5)

                logger.info("Loaders update fully concluded with success!")
        except NotInsideUpdateWindow:
            settings.set_update_control(self.update_name, "$NOT_INSIDE_UPDATE_WINDOW")
            logger.info("Sleeping until enter the update window")
            raise
        except ErrorValidatingOpenedUsers:
            settings.set_update_control(self.update_name, "$UPDATE_LOADERS_ERRORVALIDATINGOPENEDUSERS")
            logger.info("Error validating loaders. Necessary to close all the operators and the day")
            raise
        except (Exception,) as ex:
            if rollback_necessary:
                with self.working_on_update:
                    self._rollback(backup_package_zip)

            if isinstance(ex, ErrorDownloadingLoaders):
                settings.set_update_control(self.update_name, ex.message)
            else:
                settings.set_update_control(self.update_name, "$UNKNOWN_ERROR")

            if not retry_on_error:
                self.pos_repository.mark_loader_update_as_error()

            logger.exception("Not expected exception on loaders update")
            raise
        finally:
            self._delete_directories()
            settings.set_update_control("updating", False)

    def download_and_validate_loaders(self):
        settings.set_update_control(self.update_name, "$GETTING_NEW_LOADERS")
        self._download_new_loaders()

        self.updater_service.create_dir(self.work_path)
        self._extract_new_loaders(os.path.join(self.work_path, self.data_relative_path))

        settings.set_update_control(self.update_name, "$VALIDATING_NEW_LOADERS")
        current_pos_ctrl_dir = os.path.join(self.work_path, self.server_relative_path, 'bundles', 'posctrl')
        changed_pos_ctrl_type = None
        if os.path.exists(current_pos_ctrl_dir):
            changed_pos_ctrl_type = self._could_apply_new_loaders()

        return changed_pos_ctrl_type

    def _update_pos_ctrl_type(self, pos_ctrl_type):
        # type: (Dict) -> None

        for pos_id, value in pos_ctrl_type.items():
            user_control_type = 1 if value == "QSR" else 2
            self.pos_repository.update_pos_ctrl_type(pos_id, user_control_type)

    def update_operator_levels(
            self
    ):
        # type: () -> None

        try:
            if os.path.exists(self.genesis_operator_levels_json) and os.path.exists(self.sui_static_path):
                current_operator_levels = os.path.join(self.sui_static_path, 'operatorLevels.json')

                with open(self.genesis_operator_levels_json, 'rb') as f:
                    new_operator_level = json.load(f)

                has_some_configuration = sum([len(x) for x in new_operator_level["levels"].values()]) > 0
                if has_some_configuration:
                    self._create_support_menu_buttons(
                        operator_level=new_operator_level
                    )
                    self.updater_service.save_json(
                        data=new_operator_level,
                        destination_path=current_operator_levels
                    )
                else:
                    self.updater_service.delete_file(
                        path=current_operator_levels
                    )
        except (Exception,) as e:
            logger.info("Error processing operatorLevels.json: {}".format(e))
            raise ErrorUpdatingUserRoles()

    def send_result_message(
        self,                 # type: LoaderUpdaterService
        current_update_id,    # type: str
        external_update_id,   # type: str
    ):
        # type: (...) -> None

        last_versions, versions_dates = self.version_repository.get_last_versions()
        body = self._get_update_notify_body(
            last_versions=last_versions,
            versions_dates=versions_dates,
        )
        success = self.api_repository.send_result_message(
            last_versions=last_versions,
            body=body,
        )
        if success:
            self.pos_repository.update_loader_notified_date(
                update_id=current_update_id,
                external_update_id=external_update_id,
            )

    def _get_update_notify_body(
        self,               # type: LoaderUpdaterService
        last_versions,      # type: Dict[str, str]
        versions_dates,     # type: Dict[str, str]
    ):
        # type: (...) -> Dict[str, str]

        src_version = last_versions.get(UpdateType.binaries.value).split("|")[1].replace("SRC:", "")
        src_applied_date = versions_dates.get(UpdateType.binaries.value)
        src_applied_date = datetime.strptime(src_applied_date, '%d/%m/%Y %H:%M')
        loader_applied_date = datetime.strptime(versions_dates.get(UpdateType.loader.value), '%d/%m/%Y %H:%M')
        return {
            "storeId": self.store_id,
            "srcVersion": src_version,
            "srcAppliedDate": src_applied_date.isoformat(),
            "menuVersion": last_versions.get(UpdateType.catalog.value).split("_")[0],
            "loaderAppliedDate": loader_applied_date.isoformat(),
            "externalUpdateId": self.pos_repository.get_last_loader_update().server_update_identifier,
        }

    @staticmethod
    def _create_support_menu_buttons(
        operator_level,     # type: Dict
    ):
        # type: (...) -> None

        rows = int(operator_level["rows"])
        support_menu = {
            "function": "SUPPORT_MENU",
            "row": rows,
            "column": 0,
            "color": "#274E13"
        }
        update_users = {
            "function": "UPDATE_USERS",
            "row": rows,
            "column": 1,
            "color": "#274E13"
        }

        operator_level["levels"]["30"] = operator_level["levels"]["10"][:]
        operator_level["levels"]["30"].extend([support_menu, update_users])

    @staticmethod
    def _add_value_to_pos_dict(
        element_name,   # type: str
        xml_path,       # type: str
        pos_dict        # type: Dict
    ):
        # type: (...) -> None

        xml_loader = eTree.parse(xml_path)
        element_pos = xml_loader.find(".//*[@name='POS']/group")
        pos = element_pos.get("name") if element_pos is not None else None
        element = xml_loader.find(".//*[@name='{}']/string".format(element_name))
        element_value = element.text if element is not None else None
        if pos and element_value:
            pos_dict[pos] = element_value

    def _validate_update_pos_values(
        self,                   # type: LoaderUpdaterService
        boh_pos_values,         # type: Dict
        current_pos_values,     # type: Dict
        key_index,              # type: str
        validate_state=True     # type: bool
    ):
        # type: (...) -> Dict

        update_pos = {}
        some_pos_has_day_opened = False
        has_new_pos = len(boh_pos_values) > len(current_pos_values)
        for pos, value in current_pos_values.items():

            pos_id = int(pos[3:])
            if pos_id > 12:
                continue

            if not has_new_pos and (pos not in boh_pos_values.keys() or value == boh_pos_values[pos]):
                continue

            pos_id = int(pos[3:])
            model = self.pos_repository.get_model(pos_id=pos_id)
            self._validate_users(model=model)
            day_is_opened_in_pos = self._day_is_opened_in_pos(model=model)
            some_pos_has_day_opened |= day_is_opened_in_pos
            if validate_state and day_is_opened_in_pos:
                logger.exception("Error, trying to update loader with day not closed")
                raise ErrorValidatingOpenedUsers("ERRORDAYISOPEN")

            logger.info("{}: has a different values for {}".format(pos, key_index))
            update_pos[pos_id] = value

        if has_new_pos and some_pos_has_day_opened:
            logger.exception("Error, trying to add new POS with day not closed")
            raise ErrorValidatingOpenedUsers("ERRORDAYISOPEN")

        return update_pos

    @staticmethod
    def _validate_users(
        model  # type: eTree.XML
    ):
        # type: (...) -> None

        operators = model.find("Operators").findall("Operator")
        for operator in operators:
            status = operator.get("state")
            if status != "LOGGEDOUT":
                logger.exception("can't update loader with operator status: {}".format(status))
                raise ErrorValidatingOpenedUsers("NEED_TO_CLOSE_OPENED_USERS")

    def _could_apply_new_loaders(self):
        # type: () -> Dict

        boh_user_control_type = {}
        current_user_control_type = {}
        boh_pod_type = {}
        current_pod_type = {}
        boh_pos_ctrl_dir = os.path.join(self.work_path, self.server_relative_path, 'bundles', 'posctrl')
        current_pos_ctrl_dir = os.path.join(self.server_path, 'bundles', 'posctrl')

        if not os.path.exists(current_pos_ctrl_dir) or not os.path.exists(boh_pos_ctrl_dir):
            return {}

        boh_loaders_paths = os.listdir(boh_pos_ctrl_dir) or []
        boh_loaders_paths = [x for x in boh_loaders_paths if ".cfg" in x]
        for path in boh_loaders_paths:
            self._add_value_to_pos_dict(
                element_name="userControlType",
                xml_path=os.path.join(boh_pos_ctrl_dir, path),
                pos_dict=boh_user_control_type
            )
            self._add_value_to_pos_dict(
                element_name="pod",
                xml_path=os.path.join(boh_pos_ctrl_dir, path),
                pos_dict=boh_pod_type
            )

        current_loaders_paths = os.listdir(current_pos_ctrl_dir) or []
        current_loaders_paths = [x for x in current_loaders_paths if ".cfg" in x]
        for path in current_loaders_paths:
            self._add_value_to_pos_dict(
                element_name="userControlType",
                xml_path=os.path.join(current_pos_ctrl_dir, path),
                pos_dict=current_user_control_type
            )
            self._add_value_to_pos_dict(
                element_name="pod",
                xml_path=os.path.join(current_pos_ctrl_dir, path),
                pos_dict=current_pod_type
            )

        self._validate_pod_type(
            boh_pod_type=boh_pod_type,
            current_pod_type=current_pod_type
        )
        return self._validate_user_control_type(
            boh_pos_values=boh_user_control_type,
            current_pos_values=current_user_control_type
        )

    def _validate_user_control_type(
        self,                   # type: LoaderUpdaterService
        boh_pos_values,         # type: Dict
        current_pos_values,     # type: Dict
    ):
        # type: (...) -> Dict

        return self._validate_update_pos_values(
            boh_pos_values=boh_pos_values,
            current_pos_values=current_pos_values,
            key_index="user_control_type"
        )

    def _validate_pod_type(
        self,                       # type: LoaderUpdaterService
        boh_pod_type,               # type: Dict
        current_pod_type,       # type: Dict
    ):
        # type: (...) -> None

        self._validate_update_pos_values(
            boh_pos_values=boh_pod_type,
            current_pos_values=current_pod_type,
            key_index="pod_type",
            validate_state=False
        )

    def _delete_directories(self):
        self.updater_service.delete_directory(self.work_path)

    def _check_update_window(self):
        # type: () -> None

        start_time = self.local_configs.window.start_time
        end_time = self.local_configs.window.end_time
        if not ValidationUtil().is_inside_allowed_window(start_time, end_time):
            str_start_time = start_time.strftime("%H:%M:%S")
            str_end_time = end_time.strftime("%H:%M:%S")
            message = "Not inside allowed window. Start: {} / End: {}".format(str_start_time, str_end_time)
            logger.info(message)
            raise NotInsideUpdateWindow()

    def _download_new_loaders(self):
        # type: () -> None

        try:
            logger.info("Downloading new loaders")
            file_path = os.path.join(self.downloads_path, self.loaders_zip)
            self.updater_service.create_dir(path=self.downloads_path)
            last_versions, _ = self.version_repository.get_last_versions()
            self.api_repository.download_loaders(
                file_path=file_path,
                last_versions=last_versions
            )
            logger.info("Loaders successfully downloaded")
        except (Exception,) as ex:
            logger.exception("Error downloading loaders")
            raise ErrorDownloadingLoaders(ex.message)

    def _create_backup(self):
        # type: () -> str

        try:
            logger.info("Creating loaders backup")

            backup_zip_name = "loaders_{}".format(datetime.now().strftime("%Y%m%d_%H%M%S"))
            backup_package = os.path.join(self.backups_path, backup_zip_name)
            backup_package_zip = os.path.join(self.backups_path, backup_zip_name) + ".zip"

            zip_path = os.path.join(self.work_path, self.data_relative_path)
            self.updater_service.copy_directory(self.data_path, zip_path)
            self.updater_service.save_directory_zip(zip_path, backup_package)
            self.updater_service.test_zip_file(backup_package_zip)
            self._clean_backups()

            logger.info("Loaders backup successfully created!")
        except (Exception,):
            logger.exception("Error creating loaders backup")
            raise ErrorCreatingBackup()

        return backup_package_zip

    def _clean_backups(
        self,   # type: LoaderUpdaterService
    ):
        # type: (...) -> None

        list_of_files = os.listdir(self.backups_path)
        if len(list_of_files) > 5:
            full_path_files = [os.path.join(self.backups_path, x) for x in list_of_files]
            full_path_files = sorted(full_path_files, key=lambda p: os.path.getmtime(p))
            files_to_delete = full_path_files[:-5]
            for backup_file in files_to_delete:
                try:
                    os.remove(backup_file)
                except (Exception,) as e:
                    logger.exception("Failed to remove backup file {}. Error {}".format(backup_file, e))

    def _delete_content(self, path):
        # type: (str) -> None

        try:
            logger.info("Deleting path {}".format(path))

            self.updater_service.delete_directory(path)

            logger.info("Path deleted")
        except (Exception,):
            logger.exception("Error deleting path")
            raise ErrorDeletingContent()

    def _extract_new_loaders(self, path):
        # type: (str) -> None

        try:
            logger.info("Extracting loaders in data path")

            self.updater_service.create_dir(path)
            zip_path = os.path.join(self.downloads_path, self.loaders_zip)
            self.updater_service.extract_zip(zip_path, path)

            logger.info("Loaders extracted in data path")
        except (Exception,):
            logger.exception("Error extracting loaders to data path")
            raise ErrorExtractingLoaders()

    def _recover_folders(self):
        # type: () -> None

        current_folder = None
        try:
            work_path_server_folder = os.path.join(self.work_path, self.server_relative_path)
            if not os.path.exists(work_path_server_folder):
                return

            for current_folder in os.listdir(work_path_server_folder):
                if current_folder == "bundles":
                    continue

                logger.info("Recovering {} folder".format(current_folder))
                folder_path = os.path.join(self.server_path, current_folder)
                work_path_path = os.path.join(work_path_server_folder, current_folder)
                if not os.path.isdir(work_path_path):
                    continue

                self.updater_service.copy_directory(work_path_path, folder_path)
                logger.info("{} folder recovered".format(current_folder))
        except (Exception,):
            logger.exception("Error recovering {}".format(current_folder))
            raise ErrorRecoveringFolder()

    def _rollback(self, backup_package_zip):
        # type: (str) -> None

        logger.info("Starting rollback")

        retries = 0
        max_retries = 2

        while True:
            try:
                self._delete_content(self.data_path)
                self.updater_service.extract_zip(backup_package_zip, self.data_path)
            except (Exception,):
                logger.exception("Error while executing rollback")
                if retries == max_retries:
                    raise RollbackError()

                retries += 1
                continue

            logger.info("Rollback finished")
            break

    def _update_ui_customizations(
        self,
    ):
        # (...) -> None

        try:
            customization_kui_dir = os.path.join(
                self.server_path,
                "bundles",
                "customization",
                "kui",
            )
            if not os.path.isdir(customization_kui_dir):
                return

            self.updater_service.create_dir(
                path=self.htdocs_kui_path,
                delete_if_exists=False,
            )

            os_walk = os.walk(customization_kui_dir)
            for step_index, (base_dir, dirs, files) in enumerate(os_walk):
                for dir_name in dirs:
                    file_path = os.path.join(self.htdocs_kui_path, dir_name)
                    self.updater_service.create_dir(
                        path=file_path,
                        delete_if_exists=False,
                    )

                last_path_component = os.path.split(base_dir)[1]
                use_last_path_component = step_index != 0
                for file_name in files:
                    src_file = os.path.join(
                        base_dir,
                        file_name,
                    )
                    if use_last_path_component:
                        dst_file = os.path.join(
                            self.htdocs_kui_path,
                            last_path_component,
                            file_name,
                        )
                    else:
                        dst_file = os.path.join(
                            self.htdocs_kui_path,
                            file_name,
                        )
                    shutil.copy2(
                        src_file,
                        dst_file,
                    )
        except (Exception,):
            logger.exception("Error updating customization files")

    @staticmethod
    def _day_is_opened_in_pos(
        model   # type: eTree.XML
    ):
        # type: (...) -> bool

        pos_state = model.find("PosState").get("state")
        return pos_state not in ("CLOSED", "UNDEFINED")



from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.sdk.resources import (
    SERVICE_NAME,
    Resource,
)

from ._OpenTelemetryProcessorCallback import OpenTelemetryProcessorCallback


def configure_open_telemetry(
    service_name: str,
    otlp_endpoint: str = "localhost:4317",
    batch_delay_millis: int = 5000,
):
    resource = Resource(
        attributes={
            SERVICE_NAME: service_name
        }
    )

    provider = TracerProvider(resource=resource)
    otlp_exporter = OTLPSpanExporter(
        endpoint=otlp_endpoint,
        insecure=True
    )
    processor = BatchSpanProcessor(
        span_exporter=otlp_exporter,
        schedule_delay_millis=batch_delay_millis,
    )
    provider.add_span_processor(processor)
    trace.set_tracer_provider(provider)



import os


try:
    from test.support import import_helper  # type: ignore
except ImportError:
    # Python 3.9 and earlier
    class import_helper:  # type: ignore
        from test.support import (
            modules_setup,
            modules_cleanup,
            DirsOnSysPath,
            CleanImport,
        )


try:
    from test.support import os_helper  # type: ignore
except ImportError:
    # Python 3.9 compat
    class os_helper:  # type:ignore
        from test.support import temp_dir


try:
    # Python 3.10
    from test.support.os_helper import unlink
except ImportError:
    from test.support import unlink as _unlink

    def unlink(target):
        return _unlink(os.fspath(target))



# -*- coding: utf-8 -*-
import base64
import json

from model import (
    PaymentResponseData,
    StartPaymentData,
    FinishPaymentData,
)
from paymentrouter import PaymentError

from ._CommonProcessor import _CommonProcessor


class StartPaymentProcessor(_CommonProcessor):
    def parse_data(self, data, event_type):
        return StartPaymentData(*data.split(';'))

    def call_business(self, obj):
        # type: (StartPaymentData) -> str
        try:
            processor = self.payment_router.get_payment_processor(obj.electronic_type)
            add_payment_response = processor.add_payment(obj)

            if obj.tender_id and hasattr(add_payment_response, 'id_auth'):
                self.payment_router.finish_all_payments(
                    FinishPaymentData(
                        pos_id=obj.pos_id,
                        order_id=obj.order_id,
                        fiscal_date=obj.fiscal_date,
                        fiscal_time=obj.fiscal_time,
                        status="1",
                        tender_details=json.dumps(base64.b64encode(self.get_tender_details(add_payment_response))),
                        tender_id=obj.tender_id,
                    )
                )

            return add_payment_response.to_xml()
        except PaymentError:
            self.payment_router.logger.warning("Error in Start Payment Processor", exec_info=True)

    @staticmethod
    def get_tender_details(
        payment_data  # type: PaymentResponseData
    ):
        # type: (...) -> str

        tender_details = json.dumps({
            "Media": payment_data.media.strip(),
            "CNPJAuth": payment_data.cnpj_auth,
            "TransactionProcessor": payment_data.transaction_processor,
            "Bandeira": payment_data.bandeira,
            "IdAuth": payment_data.id_auth,
            "AuthCode": payment_data.auth_code,
            "ReceiptCustomer": payment_data.receipt_customer,
            "ReceiptMerchant": payment_data.receipt_merchant,
            "AccreditorSat": payment_data.sat_acquirer if hasattr(payment_data, 'sat_acquirer') else None,
            "tBand": payment_data.t_band if hasattr(payment_data, 't_band') else None,
            "tPag": payment_data.t_pag if hasattr(payment_data, 't_pag') else None,
            "NsuHost": payment_data.nsu_host if hasattr(payment_data, 'nsu_host') else None,
            "CardNumber": payment_data.card_number if hasattr(payment_data, 'card_number') else None,
            "TransactionIdentifier": payment_data.transaction_identifier,
        })

        return tender_details



from logging import Logger

from messagebus import MessageBus
from production.manager import ProductionManager
from production.model import PrinterNameFoundException
from production.treebuilder import ViewConfiguration
from production.view.printerviewservices.reportgenerator import PickListReportGenerator
from production.view.printerviewservices.reportprinter import PickListPrinter
from typing import Dict
from ._CommonViewBuilder import get_extra, get_logger

from production.view import PrinterView


class PrinterViewBuilder(object):
    def __init__(
        self,                           # type: PrinterViewBuilder
        message_bus,                    # type: MessageBus
        loggers,                        # type: Dict[str, Logger]
        production_manager,             # type: ProductionManager
        timeout_to_cleanup_devices,     # type: int
    ):
        # type: (...) -> None

        self.production_manager = production_manager
        self.message_bus = message_bus
        self.loggers = loggers
        self.timeout_to_cleanup_devices = timeout_to_cleanup_devices

    def build(self, config):
        # type: (ViewConfiguration) -> PrinterView

        logger = get_logger(config, self.loggers)

        if "PrinterName" not in config.extra_config:
            raise PrinterNameFoundException("PrinterView without printer: {}".format(config.name))

        picklist_report_generator = PickListReportGenerator(
            message_bus=self.message_bus,
            configuration_json_path=config.extra_config.get("ConfigurationJsonPath", "").encode("utf-8"),
            picklist_name=config.extra_config.get("PickListName", "").encode("utf-8"),
            logger=logger
        )

        picklist_printer = PickListPrinter(
            message_bus=self.message_bus,
            timeout=int(get_extra(config, "Timeout", 10000)) * 1000,
            name=config.name,
            logger=logger,
            kds_id=str(config.extra_config.get("KdsId", "").replace("KDS", "") or ""),
            printer=config.extra_config["PrinterName"].encode("utf-8"),
            timeout_to_cleanup_devices=self.timeout_to_cleanup_devices,
        )

        return PrinterView(
            name=config.name,
            message_bus=self.message_bus,
            report=config.extra_config["ReportName"].encode("utf-8"),
            printer=config.extra_config["PrinterName"].encode("utf-8"),
            picklist_name=config.extra_config.get("PickListName", "").encode("utf-8"),
            timeout=int(get_extra(config, "Timeout", 10000)) * 1000,
            production_manager=self.production_manager,
            report_generator=picklist_report_generator,
            picklist_printer=picklist_printer,
            tags_on_print=get_extra(config, "TagOnPrint", "printed").split(";"),
            logger=get_logger(config, self.loggers),
        )



# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

# Copyright (C) 2004-2007, 2009-2011, 2016 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose with or without fee is hereby granted,
# provided that the above copyright notice and this permission notice
# appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import struct

import dns.exception
import dns.rdata
import dns.rdatatype
import dns.name
from dns._compat import xrange

class CSYNC(dns.rdata.Rdata):

    """CSYNC record

    @ivar serial: the SOA serial number
    @type serial: int
    @ivar flags: the CSYNC flags
    @type flags: int
    @ivar windows: the windowed bitmap list
    @type windows: list of (window number, string) tuples"""

    __slots__ = ['serial', 'flags', 'windows']

    def __init__(self, rdclass, rdtype, serial, flags, windows):
        super(CSYNC, self).__init__(rdclass, rdtype)
        self.serial = serial
        self.flags = flags
        self.windows = windows

    def to_text(self, origin=None, relativize=True, **kw):
        text = ''
        for (window, bitmap) in self.windows:
            bits = []
            for i in xrange(0, len(bitmap)):
                byte = bitmap[i]
                for j in xrange(0, 8):
                    if byte & (0x80 >> j):
                        bits.append(dns.rdatatype.to_text(window * 256 +
                                                          i * 8 + j))
            text += (' ' + ' '.join(bits))
        return '%d %d%s' % (self.serial, self.flags, text)

    @classmethod
    def from_text(cls, rdclass, rdtype, tok, origin=None, relativize=True):
        serial = tok.get_uint32()
        flags = tok.get_uint16()
        rdtypes = []
        while 1:
            token = tok.get().unescape()
            if token.is_eol_or_eof():
                break
            nrdtype = dns.rdatatype.from_text(token.value)
            if nrdtype == 0:
                raise dns.exception.SyntaxError("CSYNC with bit 0")
            if nrdtype > 65535:
                raise dns.exception.SyntaxError("CSYNC with bit > 65535")
            rdtypes.append(nrdtype)
        rdtypes.sort()
        window = 0
        octets = 0
        prior_rdtype = 0
        bitmap = bytearray(b'\0' * 32)
        windows = []
        for nrdtype in rdtypes:
            if nrdtype == prior_rdtype:
                continue
            prior_rdtype = nrdtype
            new_window = nrdtype // 256
            if new_window != window:
                windows.append((window, bitmap[0:octets]))
                bitmap = bytearray(b'\0' * 32)
                window = new_window
            offset = nrdtype % 256
            byte = offset // 8
            bit = offset % 8
            octets = byte + 1
            bitmap[byte] = bitmap[byte] | (0x80 >> bit)

        windows.append((window, bitmap[0:octets]))
        return cls(rdclass, rdtype, serial, flags, windows)

    def to_wire(self, file, compress=None, origin=None):
        file.write(struct.pack('!IH', self.serial, self.flags))
        for (window, bitmap) in self.windows:
            file.write(struct.pack('!BB', window, len(bitmap)))
            file.write(bitmap)

    @classmethod
    def from_wire(cls, rdclass, rdtype, wire, current, rdlen, origin=None):
        if rdlen < 6:
            raise dns.exception.FormError("CSYNC too short")
        (serial, flags) = struct.unpack("!IH", wire[current: current + 6])
        current += 6
        rdlen -= 6
        windows = []
        while rdlen > 0:
            if rdlen < 3:
                raise dns.exception.FormError("CSYNC too short")
            window = wire[current]
            octets = wire[current + 1]
            if octets == 0 or octets > 32:
                raise dns.exception.FormError("bad CSYNC octets")
            current += 2
            rdlen -= 2
            if rdlen < octets:
                raise dns.exception.FormError("bad CSYNC bitmap length")
            bitmap = bytearray(wire[current: current + octets].unwrap())
            current += octets
            rdlen -= octets
            windows.append((window, bitmap))
        return cls(rdclass, rdtype, serial, flags, windows)



#
# The Python Imaging Library
# $Id$
#
# screen grabber (macOS and Windows only)
#
# History:
# 2001-04-26 fl  created
# 2001-09-17 fl  use builtin driver, if present
# 2002-11-19 fl  added grabclipboard support
#
# Copyright (c) 2001-2002 by Secret Labs AB
# Copyright (c) 2001-2002 by Fredrik Lundh
#
# See the README file for information on usage and redistribution.
#

import sys

from . import Image

if sys.platform == "win32":
    grabber = Image.core.grabscreen
elif sys.platform == "darwin":
    import os
    import tempfile
    import subprocess
else:
    raise ImportError("ImageGrab is macOS and Windows only")


def grab(bbox=None, include_layered_windows=False, all_screens=False):
    if sys.platform == "darwin":
        fh, filepath = tempfile.mkstemp(".png")
        os.close(fh)
        subprocess.call(["screencapture", "-x", filepath])
        im = Image.open(filepath)
        im.load()
        os.unlink(filepath)
        if bbox:
            im = im.crop(bbox)
    else:
        offset, size, data = grabber(include_layered_windows, all_screens)
        im = Image.frombytes(
            "RGB",
            size,
            data,
            # RGB, 32-bit line padding, origin lower left corner
            "raw",
            "BGR",
            (size[0] * 3 + 3) & -4,
            -1,
        )
        if bbox:
            x0, y0 = offset
            left, top, right, bottom = bbox
            im = im.crop((left - x0, top - y0, right - x0, bottom - y0))
    return im


def grabclipboard():
    if sys.platform == "darwin":
        fh, filepath = tempfile.mkstemp(".jpg")
        os.close(fh)
        commands = [
            'set theFile to (open for access POSIX file "'
            + filepath
            + '" with write permission)',
            "try",
            "    write (the clipboard as JPEG picture) to theFile",
            "end try",
            "close access theFile",
        ]
        script = ["osascript"]
        for command in commands:
            script += ["-e", command]
        subprocess.call(script)

        im = None
        if os.stat(filepath).st_size != 0:
            im = Image.open(filepath)
            im.load()
        os.unlink(filepath)
        return im
    else:
        data = Image.core.grabclipboard()
        if isinstance(data, bytes):
            from . import BmpImagePlugin
            import io

            return BmpImagePlugin.DibImageFile(io.BytesIO(data))
        return data



# -*- coding: utf-8 -*-
import json

import sysactions
from actions.util import get_ruptured_products_from_remote_order

from .. import logger


@sysactions.action
def get_ruptured_products(pos_id):
    # type: (str) -> str

    try:
        products = {
            "enabledProducts": [],
            "disabledProducts": get_ruptured_products_from_remote_order()
        }

        return json.dumps(products)

    except Exception as e:
        logger.exception('[get_ruptured_products] Exception {}'.format(e))
        sysactions.show_messagebox(pos_id, "$RUPTURE_ERROR", "$INFORMATION")



# nhs.py - functions for handling United Kingdom NHS numbers
#
# Copyright (C) 2016-2017 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""NHS (United Kingdom National Health Service patient identifier).

The NHS number is used by the NHS (National Health Service) and its partners
to uniquely identify patients. The number is used in England, Wales and the
Isle of Man. The number is assigned at birth and consists of 10 digits where
the final digit is a check digit.

More information:

* https://en.wikipedia.org/wiki/NHS_number
* https://www.nhs.uk/using-the-nhs/about-the-nhs/what-is-an-nhs-number/
* https://digital.nhs.uk/article/301/NHS-Number
* https://www.datadictionary.nhs.uk/data_dictionary/attributes/n/nhs/nhs_number_de.asp

>>> validate('943-476-5870')
'9434765870'
>>> validate('9434765871')  # invalid check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> format('9434765870')
'943 476 5870'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').strip()


def checksum(number):
    """Calculate the checksum. The checksum is only used for the 9 digits
    of the number and the result can either be 0 or 42."""
    return sum(i * int(n) for i, n in enumerate(reversed(number), 1)) % 11


def validate(number):
    """Check if the number is valid. This checks the length and check
    digit."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) != 10:
        raise InvalidLength()
    if checksum(number) != 0:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is valid."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number, separator=' '):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return separator.join((number[0:3], number[3:6], number[6:]))



from production.box._OrderChangerProductionBox import OrderChangerProductionBox


class OrderStampBox(OrderChangerProductionBox):
    def __init__(self, name, sons, logger=None):
        super(OrderStampBox, self).__init__(name, sons, logger)

    def change_order(self, order):
        order.stamped = True
        for item in order.items:
            item.stamped = True
        return order



def migrate(conn, tables):
    update_tables_priority = [
        "Event",
        "PartCodeQuantity",
    ]

    for table in update_tables_priority:
        if len(tables[table]["fields_Common"]) == 0:
            continue

        str_common_fields = ", ".join(tables[table]["fields_Common"])
        insert_str = "INSERT OR REPLACE INTO {} ({}) SELECT {} FROM old.{};"
        stmt = insert_str.format(table, str_common_fields, str_common_fields, table)
        conn.query(stmt)



from __future__ import absolute_import, division, print_function, unicode_literals

from base64 import b64encode, b64decode
from enum import Enum

from eight import str, bytes
from lxml import etree
from lxml.etree import Element, SubElement
from defusedxml.lxml import fromstring

from cryptography.hazmat.primitives.asymmetric import rsa, dsa, ec
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.hashes import Hash, SHA1, SHA224, SHA256, SHA384, SHA512
from cryptography.hazmat.backends import default_backend

from pyasn1.codec.der import encoder as der_encoder, decoder as der_decoder

from .exceptions import InvalidSignature, InvalidDigest, InvalidInput, InvalidCertificate  # noqa
from .util import (bytes_to_long, long_to_bytes, strip_pem_header, add_pem_header, ensure_bytes, ensure_str, Namespace,
                   XMLProcessor, DERSequenceOfIntegers, iterate_pem, verify_x509_cert_chain)
from collections import namedtuple

methods = Enum("Methods", "enveloped enveloping detached")

namespaces = Namespace(
    ds="http://www.w3.org/2000/09/xmldsig#",
    dsig11="http://www.w3.org/2009/xmldsig11#",
    dsig2="http://www.w3.org/2010/xmldsig2#",
    ec="http://www.w3.org/2001/10/xml-exc-c14n#",
    dsig_more="http://www.w3.org/2001/04/xmldsig-more#",
    xenc="http://www.w3.org/2001/04/xmlenc#",
    xenc11="http://www.w3.org/2009/xmlenc11#"
)

def ds_tag(tag):
    return "{" + namespaces.ds + "}" + tag

def dsig11_tag(tag):
    return "{" + namespaces.dsig11 + "}" + tag

def _remove_sig(signature, idempotent=False):
    """
    Remove the signature node from its parent, keeping any tail element.
    This is needed for eneveloped signatures.

    :param signature: Signature to remove from payload
    :type signature: XML ElementTree Element
    :param idempotent:
        If True, don't raise an error if signature is already detached from parent.
    :type idempotent: boolean
    """
    try:
        signaturep = next(signature.iterancestors())
    except StopIteration:
        if idempotent:
            return
        raise ValueError("Can't remove the root signature node")
    if signature.tail is not None:
        try:
            signatures = next(signature.itersiblings(preceding=True))
        except StopIteration:
            if signaturep.text is not None:
                signaturep.text = signaturep.text + signature.tail
            else:
                signaturep.text = signature.tail
        else:
            if signatures.tail is not None:
                signatures.tail = signatures.tail + signature.tail
            else:
                signatures.tail = signature.tail
    signaturep.remove(signature)

class VerifyResult(namedtuple("VerifyResult", "signed_data signed_xml signature_xml")):
    """
    The results of a verification return the signed data, the signed xml and the signature xml

    :param signed_data: The binary data as it was signed (literally)
    :type data: bytes
    :param signed_xml: The signed data parsed as XML (or None if parsing failed)
    :type signed_xml: ElementTree or None
    :param signature_xml: The signature element parsed as XML
    :type signed_xml: ElementTree

    This class is a namedtuple representing structured data returned by ``signxml.XMLVerifier.verify()``. As with any
    namedtuple, elements of the return value can be accessed as attributes. For example::
        verified_data = signxml.XMLVerifier().verify(input_data).signed_xml
    """

class XMLSignatureProcessor(XMLProcessor):
    schema_file = "xmldsig1-schema.xsd"

    known_digest_methods = {
        namespaces.ds + "sha1": SHA1,
        namespaces.xenc + "sha256": SHA256,
        namespaces.dsig_more + "sha224": SHA224,
        namespaces.dsig_more + "sha384": SHA384,
        namespaces.xenc + "sha512": SHA512,
    }

    known_hmac_digest_methods = {
        namespaces.ds + "hmac-sha1": SHA1,
        namespaces.dsig_more + "hmac-sha256": SHA256,
        namespaces.dsig_more + "hmac-sha384": SHA384,
        namespaces.dsig_more + "hmac-sha512": SHA512,
        namespaces.dsig_more + "hmac-sha224": SHA224,
    }

    known_signature_digest_methods = {
        namespaces.dsig_more + "rsa-sha256": SHA256,
        namespaces.dsig_more + "ecdsa-sha256": SHA256,
        namespaces.ds + "dsa-sha1": SHA1,
        namespaces.ds + "rsa-sha1": SHA1,
        namespaces.dsig_more + "rsa-sha224": SHA224,
        namespaces.dsig_more + "rsa-sha384": SHA384,
        namespaces.dsig_more + "rsa-sha512": SHA512,
        namespaces.dsig_more + "ecdsa-sha1": SHA1,
        namespaces.dsig_more + "ecdsa-sha224": SHA224,
        namespaces.dsig_more + "ecdsa-sha384": SHA384,
        namespaces.dsig_more + "ecdsa-sha512": SHA512,
        namespaces.dsig11 + "dsa-sha256": SHA256,
    }
    known_digest_tags = {method.split("#")[1]: method for method in known_digest_methods}
    known_hmac_digest_tags = {method.split("#")[1]: method for method in known_hmac_digest_methods}
    known_signature_digest_tags = {method.split("#")[1]: method for method in known_signature_digest_methods}

    # See https://tools.ietf.org/html/rfc5656
    known_ecdsa_curves = {
        "urn:oid:1.2.840.10045.3.1.7": ec.SECP256R1,
        "urn:oid:1.3.132.0.34": ec.SECP384R1,
        "urn:oid:1.3.132.0.35": ec.SECP521R1,
        "urn:oid:1.3.132.0.1": ec.SECT163K1,
        "urn:oid:1.2.840.10045.3.1.1": ec.SECP192R1,
        "urn:oid:1.3.132.0.33": ec.SECP224R1,
        "urn:oid:1.3.132.0.26": ec.SECT233K1,
        "urn:oid:1.3.132.0.27": ec.SECT233R1,
        "urn:oid:1.3.132.0.16": ec.SECT283R1,
        "urn:oid:1.3.132.0.36": ec.SECT409K1,
        "urn:oid:1.3.132.0.37": ec.SECT409R1,
        "urn:oid:1.3.132.0.38": ec.SECT571K1,
    }
    known_ecdsa_curve_oids = {ec().name: oid for oid, ec in known_ecdsa_curves.items()}

    known_c14n_algorithms = {
        "http://www.w3.org/TR/2001/REC-xml-c14n-20010315",
        "http://www.w3.org/TR/2001/REC-xml-c14n-20010315#WithComments",
        "http://www.w3.org/2001/10/xml-exc-c14n#",
        "http://www.w3.org/2001/10/xml-exc-c14n#WithComments",
        "http://www.w3.org/2006/12/xml-c14n11",
        "http://www.w3.org/2006/12/xml-c14n11#WithComments"
    }
    default_c14n_algorithm = "http://www.w3.org/2006/12/xml-c14n11"

    id_attributes = ("Id", "ID", "id", "xml:id")

    def _get_digest(self, data, digest_algorithm):
        hasher = Hash(algorithm=digest_algorithm, backend=default_backend())
        hasher.update(data)
        return ensure_str(b64encode(hasher.finalize()))

    def _get_digest_method(self, digest_algorithm_id, methods=None):
        if methods is None:
            methods = self.known_digest_methods
        if digest_algorithm_id not in methods:
            raise InvalidInput('Algorithm "{}" is not recognized'.format(digest_algorithm_id))
        return methods[digest_algorithm_id]()

    def _get_digest_method_by_tag(self, digest_algorithm_tag, methods=None, known_tags=None):
        if known_tags is None:
            known_tags = self.known_digest_tags
        if digest_algorithm_tag not in known_tags:
            raise InvalidInput('Algorithm tag "{}" is not recognized'.format(digest_algorithm_tag))
        return self._get_digest_method(known_tags[digest_algorithm_tag], methods=methods)

    def _get_hmac_digest_method(self, hmac_algorithm_id):
        return self._get_digest_method(hmac_algorithm_id, methods=self.known_hmac_digest_methods)

    def _get_hmac_digest_method_by_tag(self, hmac_algorithm_tag):
        return self._get_digest_method_by_tag(hmac_algorithm_tag, methods=self.known_hmac_digest_methods,
                                              known_tags=self.known_hmac_digest_tags)

    def _get_signature_digest_method(self, signature_algorithm_id):
        return self._get_digest_method(signature_algorithm_id, methods=self.known_signature_digest_methods)

    def _get_signature_digest_method_by_tag(self, signature_algorithm_tag):
        return self._get_digest_method_by_tag(signature_algorithm_tag, methods=self.known_signature_digest_methods,
                                              known_tags=self.known_signature_digest_tags)

    def _find(self, element, query, require=True, namespace="ds", anywhere=False):
        if anywhere:
            result = element.find('.//' + namespace + ":" + query, namespaces=namespaces)
        else:
            result = element.find(namespace + ":" + query, namespaces=namespaces)

        if require and result is None:
            raise InvalidInput("Expected to find XML element {} in {}".format(query, element.tag))
        return result

    def _findall(self, element, query, namespace="ds", anywhere=False):
        if anywhere:
            return element.findall('.//' + namespace + ":" + query, namespaces=namespaces)
        else:
            return element.findall(namespace + ":" + query, namespaces=namespaces)

    def _c14n(self, nodes, algorithm, inclusive_ns_prefixes=None):
        exclusive, with_comments = False, False

        if algorithm.startswith("http://www.w3.org/2001/10/xml-exc-c14n#"):
            exclusive = True
        if algorithm.endswith("#WithComments"):
            with_comments = True

        if not isinstance(nodes, list):
            nodes = [nodes]

        c14n = b""
        for node in nodes:
            c14n += etree.tostring(node, method="c14n", exclusive=exclusive, with_comments=with_comments,
                                   inclusive_ns_prefixes=inclusive_ns_prefixes)  # TODO: optimize if needed
        if exclusive is False:
            # TODO: there must be a nicer way to do this. See also:
            # http://www.w3.org/TR/xml-c14n, "namespace axis"
            # http://www.w3.org/TR/xml-c14n2/#sec-Namespace-Processing
            c14n = c14n.replace(b' xmlns=""', b'')
        return c14n

    def _resolve_reference(self, doc_root, reference, uri_resolver=None):
        uri = reference.get("URI")
        if uri == "":
            return doc_root
        elif uri.startswith("#xpointer("):
            raise InvalidInput("XPointer references are not supported")
            # doc_root.xpath(uri.lstrip("#"))[0]
        elif uri.startswith("#"):
            for id_attribute in self.id_attributes:
                xpath_query = "//*[@*[local-name() = '{}']=$uri]".format(id_attribute)
                results = doc_root.xpath(xpath_query, uri=uri.lstrip("#"))
                if len(results) > 1:
                    raise InvalidInput("Ambiguous reference URI {} resolved to {} nodes".format(uri, len(results)))
                elif len(results) == 1:
                    return results[0]
            raise InvalidInput("Unable to resolve reference URI: {}".format(uri))
        else:
            if uri_resolver is None:
                raise InvalidInput("External URI dereferencing is not configured: {}".format(uri))
            result = uri_resolver(uri)
            if result is None:
                raise InvalidInput("Unable to resolve reference URI: {}".format(uri))
            return result

class XMLSigner(XMLSignatureProcessor):
    """
    Create a new XML Signature Signer object, which can be used to hold configuration information and sign multiple
    pieces of data.

    :param method:
        ``signxml.methods.enveloped``, ``signxml.methods.enveloping``, or ``signxml.methods.detached``. See the list
        of signature types under `XML Signature Syntax and Processing Version 2.0, Definitions
        <http://www.w3.org/TR/xmldsig-core2/#sec-Definitions>`_.
    :type method: :py:class:`methods`
    :param signature_algorithm:
        Algorithm that will be used to generate the signature, composed of the signature algorithm and the digest
        algorithm, separated by a hyphen. All algorithm IDs listed under the `Algorithm Identifiers and
        Implementation Requirements <http://www.w3.org/TR/xmldsig-core1/#sec-AlgID>`_ section of the XML Signature
        1.1 standard are supported.
    :type signature_algorithm: string
    :param digest_algorithm: Algorithm that will be used to hash the data during signature generation. All algorithm IDs
        listed under the `Algorithm Identifiers and Implementation Requirements
        <http://www.w3.org/TR/xmldsig-core1/#sec-AlgID>`_ section of the XML Signature 1.1 standard are supported.
    :type digest_algorithm: string
    """
    def __init__(self, method=methods.enveloped, signature_algorithm="rsa-sha256", digest_algorithm="sha256",
                 c14n_algorithm=XMLSignatureProcessor.default_c14n_algorithm):
        if method not in methods:
            raise InvalidInput("Unknown signature method {}".format(method))
        self.method = method
        self.sign_alg = signature_algorithm
        assert self.sign_alg in self.known_signature_digest_tags or self.sign_alg in self.known_hmac_digest_tags
        assert digest_algorithm in self.known_digest_tags
        self.digest_alg = digest_algorithm
        assert c14n_algorithm in self.known_c14n_algorithms
        self.c14n_alg = c14n_algorithm
        self.namespaces = dict(ds=namespaces.ds)
        self._parser = None

    def sign(self, data, key=None, passphrase=None, cert=None, reference_uri=None, key_name=None, key_info=None,
             id_attribute=None):
        """
        Sign the data and return the root element of the resulting XML tree.

        :param data: Data to sign
        :type data: String, file-like object, or XML ElementTree Element API compatible object
        :param key:
            Key to be used for signing. When signing with a certificate or RSA/DSA/ECDSA key, this can be a string
            containing a PEM-formatted key, or a :py:class:`cryptography.hazmat.primitives.interfaces.RSAPublicKey`,
            :py:class:`cryptography.hazmat.primitives.interfaces.DSAPublicKey`, or
            :py:class:`cryptography.hazmat.primitives.interfaces.EllipticCurvePublicKey` object. When signing with a
            HMAC, this should be a string containing the shared secret.
        :type key:
            string, :py:class:`cryptography.hazmat.primitives.interfaces.RSAPublicKey`,
            :py:class:`cryptography.hazmat.primitives.interfaces.DSAPublicKey`, or
            :py:class:`cryptography.hazmat.primitives.interfaces.EllipticCurvePublicKey` object
        :param passphrase: Passphrase to use to decrypt the key, if any.
        :type passphrase: string
        :param cert:
            X.509 certificate to use for signing. This should be a string containing a PEM-formatted certificate, or an
            array of strings or OpenSSL.crypto.X509 objects containing the certificate and a chain of intermediate
            certificates.
        :type cert: string, array of strings, or array of OpenSSL.crypto.X509 objects
        :param reference_uri:
            Custom reference URI or list of reference URIs to incorporate into the signature. When ``method`` is set to
            ``detached`` or ``enveloped``, reference URIs are set to this value and only the referenced elements are
            signed.
        :type reference_uri: string or list
        :param key_name: Add a KeyName element in the KeyInfo element that may be used by the signer to communicate a
            key identifier to the recipient. Typically, KeyName contains an identifier related to the key pair used to
            sign the message.
        :type key_name: string
        :param key_info: A custom KeyInfo element to insert in the signature. Use this to supply
            ``<wsse:SecurityTokenReference>`` or other custom key references.
        :type key_info: :py:class:`lxml.etree.Element`
        :param id_attribute:
            Name of the attribute whose value ``URI`` refers to. By default, SignXML will search for "Id", then "ID".
        :type id_attribute: string

        :returns:
            A :py:class:`lxml.etree.Element` object representing the root of the XML tree containing the signature and
            the payload data.

        To specify the location of an enveloped signature within **data**, insert a
        ``<Signature Id="placeholder"></Signature>`` element in **data**. This element will be replaced by the generated
        signature, and excised when generating the digest.
        """
        if id_attribute is not None:
            self.id_attributes = (id_attribute, )

        if isinstance(cert, (str, bytes)):
            cert_chain = list(iterate_pem(cert))
        else:
            cert_chain = cert

        if isinstance(reference_uri, (str, bytes)):
            reference_uris = [reference_uri]
        else:
            reference_uris = reference_uri

        sig_root, doc_root, c14n_inputs, reference_uris = self._unpack(data, reference_uris)
        signed_info_element, signature_value_element = self._build_sig(sig_root, reference_uris, c14n_inputs)

        if key is None:
            raise InvalidInput('Parameter "key" is required')

        signed_info_c14n = self._c14n(signed_info_element, algorithm=self.c14n_alg)
        if self.sign_alg.startswith("hmac-"):
            from cryptography.hazmat.primitives.hmac import HMAC
            signer = HMAC(key=key,
                          algorithm=self._get_hmac_digest_method_by_tag(self.sign_alg),
                          backend=default_backend())
            signer.update(signed_info_c14n)
            signature_value_element.text = ensure_str(b64encode(signer.finalize()))
            sig_root.append(signature_value_element)
        elif any(self.sign_alg.startswith(i) for i in ["dsa-", "rsa-", "ecdsa-"]):
            if isinstance(key, (str, bytes)):
                from cryptography.hazmat.primitives.serialization import load_pem_private_key
                key = load_pem_private_key(key, password=passphrase, backend=default_backend())

            hash_alg = self._get_signature_digest_method_by_tag(self.sign_alg)
            if self.sign_alg.startswith("dsa-"):
                signer = key.signer(signature_algorithm=hash_alg)
            elif self.sign_alg.startswith("ecdsa-"):
                signer = key.signer(signature_algorithm=ec.ECDSA(algorithm=hash_alg))
            elif self.sign_alg.startswith("rsa-"):
                signer = key.signer(padding=PKCS1v15(), algorithm=hash_alg)
            else:
                raise NotImplementedError()
            signer.update(signed_info_c14n)
            signature = signer.finalize()
            if self.sign_alg.startswith("dsa-"):
                # Note: The output of the DSA signer is a DER-encoded ASN.1 sequence of two DER integers.
                decoded_signature = der_decoder.decode(signature)[0]
                r = decoded_signature.getComponentByPosition(0)
                s = decoded_signature.getComponentByPosition(1)
                signature = long_to_bytes(r).rjust(32, b"\0") + long_to_bytes(s).rjust(32, b"\0")

            signature_value_element.text = ensure_str(b64encode(signature))

            if key_info is None:
                key_info = SubElement(sig_root, "KeyInfo")
                if key_name is not None:
                    keyname = SubElement(key_info, "KeyName")
                    keyname.text = key_name

                if cert_chain is None:
                    self._serialize_key_value(key, key_info)
                else:
                    x509_data = SubElement(key_info, "X509Data")
                    for cert in cert_chain:
                        x509_certificate = SubElement(x509_data, "X509Certificate")
                        if isinstance(cert, (str, bytes)):
                            x509_certificate.text = strip_pem_header(cert)
                        else:
                            from OpenSSL.crypto import dump_certificate, FILETYPE_PEM
                            x509_certificate.text = dump_certificate(FILETYPE_PEM, cert)
            else:
                sig_root.append(key_info)
        else:
            raise NotImplementedError()

        if self.method == methods.enveloping:
            for c14n_input in c14n_inputs:
                doc_root.append(c14n_input)
        if self.method == methods.enveloped:
            doc_root.append(sig_root)
            return doc_root
        else:
            return sig_root

    def _get_c14n_inputs_from_reference_uris(self, doc_root, reference_uris):
        c14n_inputs, new_reference_uris = [], []
        for reference_uri in reference_uris:
            if not reference_uri.startswith('#'):
                reference_uri = '#' + reference_uri
            c14n_inputs.append(self.get_root(self._resolve_reference(doc_root, {'URI': reference_uri})))
            new_reference_uris.append(reference_uri)
        return c14n_inputs, new_reference_uris

    def _unpack(self, data, reference_uris):
        sig_root = Element("Signature", xmlns=namespaces.ds)
        if self.method == methods.enveloped:
            if isinstance(data, (str, bytes)):
                raise InvalidInput("When using enveloped signature, **data** must be an XML element")
            doc_root = self.get_root(data)
            c14n_inputs = [self.get_root(data)]
            if reference_uris is not None:
                # Only sign the referenced element(s)
                c14n_inputs, reference_uris = self._get_c14n_inputs_from_reference_uris(doc_root, reference_uris)

            signature_placeholders = self._findall(doc_root, "Signature[@Id='placeholder']", anywhere=True)
            if len(signature_placeholders) == 0:
                pass
                #doc_root.append(sig_root)
            elif len(signature_placeholders) == 1:
                sig_root = signature_placeholders[0]
                del sig_root.attrib["Id"]
                for c14n_input in c14n_inputs:
                    placeholders = self._findall(c14n_input, "Signature[@Id='placeholder']", anywhere=True)
                    if placeholders:
                        assert len(placeholders) == 1
                        _remove_sig(placeholders[0])
            else:
                raise InvalidInput("Enveloped signature input contains more than one placeholder")

            if reference_uris is None:
                # Set default reference URIs based on signed data ID attribute values
                reference_uris = []
                for c14n_input in c14n_inputs:
                    payload_id = c14n_input.get("Id", c14n_input.get("ID"))
                    reference_uris.append("#{}".format(payload_id) if payload_id is not None else "")
        elif self.method == methods.detached:
            doc_root = self.get_root(data)
            if reference_uris is None:
                reference_uris = ["#{}".format(data.get("Id", data.get("ID", "object")))]
                c14n_inputs = [self.get_root(data)]
            try:
                c14n_inputs, reference_uris = self._get_c14n_inputs_from_reference_uris(doc_root, reference_uris)
            except InvalidInput:  # Dummy reference URI
                c14n_inputs = [self.get_root(data)]
        elif self.method == methods.enveloping:
            doc_root = sig_root
            c14n_inputs = [Element(ds_tag("Object"), nsmap=self.namespaces, Id="object")]
            if isinstance(data, (str, bytes)):
                c14n_inputs[0].text = data
            else:
                c14n_inputs[0].append(self.get_root(data))
            reference_uris = ["#object"]
        return sig_root, doc_root, c14n_inputs, reference_uris

    def _build_sig(self, sig_root, reference_uris, c14n_inputs):
        signed_info = SubElement(sig_root, "SignedInfo", xmlns=namespaces.ds)
        c14n_method = SubElement(signed_info, "CanonicalizationMethod", Algorithm=self.c14n_alg)
        if self.sign_alg.startswith("hmac-"):
            algorithm_id = self.known_hmac_digest_tags[self.sign_alg]
        else:
            algorithm_id = self.known_signature_digest_tags[self.sign_alg]
        signature_method = SubElement(signed_info, "SignatureMethod", Algorithm=algorithm_id)
        for i, reference_uri in enumerate(reference_uris):
            reference = SubElement(signed_info, "Reference", URI=reference_uri)
            if self.method == methods.enveloped:
                transforms = SubElement(reference, "Transforms")
                SubElement(transforms, "Transform", Algorithm=namespaces.ds + "enveloped-signature")
                SubElement(transforms, "Transform", Algorithm=self.c14n_alg)
            digest_method = SubElement(reference, "DigestMethod",
                                       Algorithm=self.known_digest_tags[self.digest_alg])
            digest_value = SubElement(reference, "DigestValue")
            payload_c14n = self._c14n(c14n_inputs[i], algorithm=self.c14n_alg)
            digest = self._get_digest(payload_c14n, self._get_digest_method_by_tag(self.digest_alg))
            digest_value.text = digest
        signature_value = SubElement(sig_root, "SignatureValue")
        return signed_info, signature_value

    def _serialize_key_value(self, key, key_info_element):
        key_value = SubElement(key_info_element, ds_tag("KeyValue"))
        if self.sign_alg.startswith("rsa-"):
            rsa_key_value = SubElement(key_value, ds_tag("RSAKeyValue"))
            modulus = SubElement(rsa_key_value, ds_tag("Modulus"))
            modulus.text = ensure_str(b64encode(long_to_bytes(key.public_key().public_numbers().n)))
            exponent = SubElement(rsa_key_value, ds_tag("Exponent"))
            exponent.text = ensure_str(b64encode(long_to_bytes(key.public_key().public_numbers().e)))
        elif self.sign_alg.startswith("dsa-"):
            dsa_key_value = SubElement(key_value, ds_tag("DSAKeyValue"))
            for field in "p", "q", "g", "y":
                e = SubElement(dsa_key_value, ds_tag(field.upper()))

                if field == "y":
                    key_params = key.public_key().public_numbers()
                else:
                    key_params = key.parameters().parameter_numbers()

                e.text = ensure_str(b64encode(long_to_bytes(getattr(key_params, field))))
        elif self.sign_alg.startswith("ecdsa-"):
            ec_key_value = SubElement(key_value, dsig11_tag("ECKeyValue"), nsmap=dict(dsig11=namespaces.dsig11))
            named_curve = SubElement(ec_key_value, dsig11_tag("NamedCurve"),
                                     URI=self.known_ecdsa_curve_oids[key.curve.name])
            public_key = SubElement(ec_key_value, dsig11_tag("PublicKey"))
            x = key.public_key().public_numbers().x
            y = key.public_key().public_numbers().y
            public_key.text = ensure_str(b64encode(long_to_bytes(4) + long_to_bytes(x) + long_to_bytes(y)))

class XMLVerifier(XMLSignatureProcessor):
    """
    Create a new XML Signature Verifier object, which can be used to hold configuration information and verify multiple
    pieces of data.
    """
    def _verify_signature_with_pubkey(self, signed_info_c14n, raw_signature, key_value, signature_alg):
        if "ecdsa-" in signature_alg:
            ec_key_value = self._find(key_value, "ECKeyValue", namespace="dsig11")
            named_curve = self._find(ec_key_value, "NamedCurve", namespace="dsig11")
            public_key = self._find(ec_key_value, "PublicKey", namespace="dsig11")
            key_data = b64decode(public_key.text)[1:]
            x = bytes_to_long(key_data[:len(key_data)//2])
            y = bytes_to_long(key_data[len(key_data)//2:])
            curve_class = self.known_ecdsa_curves[named_curve.get("URI")]
            key = ec.EllipticCurvePublicNumbers(x=x, y=y, curve=curve_class()).public_key(backend=default_backend())
            verifier = key.verifier(raw_signature, ec.ECDSA(self._get_signature_digest_method(signature_alg)))
        elif "dsa-" in signature_alg:
            dsa_key_value = self._find(key_value, "DSAKeyValue")
            p = self._get_long(dsa_key_value, "P")
            q = self._get_long(dsa_key_value, "Q")
            g = self._get_long(dsa_key_value, "G", require=False)
            y = self._get_long(dsa_key_value, "Y")
            pn = dsa.DSAPublicNumbers(y=y, parameter_numbers=dsa.DSAParameterNumbers(p=p, q=q, g=g))
            key = pn.public_key(backend=default_backend())
            der_seq = DERSequenceOfIntegers([bytes_to_long(raw_signature[:len(raw_signature)//2]),
                                             bytes_to_long(raw_signature[len(raw_signature)//2:])])
            sig_as_der_seq = der_encoder.encode(der_seq)
            verifier = key.verifier(sig_as_der_seq, self._get_signature_digest_method(signature_alg))
        elif "rsa-" in signature_alg:
            rsa_key_value = self._find(key_value, "RSAKeyValue")
            modulus = self._get_long(rsa_key_value, "Modulus")
            exponent = self._get_long(rsa_key_value, "Exponent")
            key = rsa.RSAPublicNumbers(e=exponent, n=modulus).public_key(backend=default_backend())
            verifier = key.verifier(raw_signature, padding=PKCS1v15(),
                                    algorithm=self._get_signature_digest_method(signature_alg))
        else:
            raise NotImplementedError()

        verifier.update(signed_info_c14n)
        verifier.verify()

    def _get_inclusive_ns_prefixes(self, transform_node):
        inclusive_namespaces = transform_node.find("./ec:InclusiveNamespaces[@PrefixList]", namespaces=namespaces)
        if inclusive_namespaces is None:
            return None
        else:
            return inclusive_namespaces.get("PrefixList").split(" ")

    def _apply_transforms(self, payload, transforms_node, signature, c14n_algorithm):
        transforms, c14n_applied = [], False
        if transforms_node is not None:
            transforms = self._findall(transforms_node, "Transform")

        for transform in transforms:
            if transform.get("Algorithm") == "http://www.w3.org/2000/09/xmldsig#enveloped-signature":
                _remove_sig(signature, idempotent=True)

        for transform in transforms:
            if transform.get("Algorithm") == "http://www.w3.org/2000/09/xmldsig#base64":
                payload = b64decode(payload.text)

        for transform in transforms:
            algorithm = transform.get("Algorithm")
            if algorithm in self.known_c14n_algorithms:
                inclusive_ns_prefixes = self._get_inclusive_ns_prefixes(transform)
                payload = self._c14n(payload, algorithm=algorithm, inclusive_ns_prefixes=inclusive_ns_prefixes)
                c14n_applied = True

        if not c14n_applied and not isinstance(payload, (str, bytes)):
            payload = self._c14n(payload, algorithm=c14n_algorithm)

        return payload

    def verify(self, data, require_x509=True, x509_cert=None, cert_subject_name=None, ca_pem_file=None, ca_path=None,
               hmac_key=None, validate_schema=True, parser=None, uri_resolver=None, id_attribute=None,
               expect_references=1):
        """
        Verify the XML signature supplied in the data and return the XML node signed by the signature, or raise an
        exception if the signature is not valid. By default, this requires the signature to be generated using a valid
        X.509 certificate. To enable other means of signature validation, set the **require_x509** argument to `False`.

        .. admonition:: See what is signed

         It is important to understand and follow the best practice rule of "See what is signed" when verifying XML
         signatures. The gist of this rule is: if your application neglects to verify that the information it trusts is
         what was actually signed, the attacker can supply a valid signature but point you to malicious data that wasn't
         signed by that signature.

         In SignXML, you can ensure that the information signed is what you expect to be signed by only trusting the
         data returned by the ``verify()`` method. The return value is the XML node or string that was signed. Also,
         depending on the signature settings used, comments in the XML data may not be subject to signing, so may need
         to be untrusted.

         **Recommended reading:** http://www.w3.org/TR/xmldsig-bestpractices/#practices-applications

        .. admonition:: Establish trust

         If you do not supply any keyword arguments to ``verify()``, the default behavior is to trust **any** valid XML
         signature generated using a valid X.509 certificate trusted by your system's CA store. This means anyone can
         get an SSL certificate and generate a signature that you will trust. To establish trust in the signer, use the
         ``x509_cert`` argument to specify a certificate that was pre-shared out-of-band (e.g. via SAML metadata, as
         shown in :ref:`Verifying SAML assertions <verifying-saml-assertions>`), or ``cert_subject_name`` to specify a
         subject name that must be in the signing X.509 certificate given by the signature (verified as if it were a
         domain name), or ``ca_pem_file``/``ca_path`` to give a custom CA.

        :param data: Signature data to verify
        :type data: String, file-like object, or XML ElementTree Element API compatible object
        :param require_x509:
            If ``True``, a valid X.509 certificate-based signature with an established chain of trust is required to
            pass validation. If ``False``, other types of valid signatures (e.g. HMAC or RSA public key) are accepted.
        :type require_x509: boolean
        :param x509_cert:
            A trusted external X.509 certificate, given as a PEM-formatted string or OpenSSL.crypto.X509 object, to use
            for verification. Overrides any X.509 certificate information supplied by the signature. If left set to
            ``None``, requires that the signature supply a valid X.509 certificate chain that validates against the
            known certificate authorities. Implies **require_x509=True**.
        :type x509_cert: string or OpenSSL.crypto.X509
        :param ca_pem_file:
            Filename of a PEM file containing certificate authority information to use when verifying certificate-based
            signatures.
        :type ca_pem_file: string or bytes
        :param ca_path:
            Path to a directory containing PEM-formatted certificate authority files to use when verifying
            certificate-based signatures. If neither **ca_pem_file** nor **ca_path** is given, the Mozilla CA bundle
            provided by :py:mod:`certifi` will be loaded.
        :type ca_path: string
        :param cert_subject_name:
            Subject Common Name to check the signing X.509 certificate against. Implies **require_x509=True**.
        :type cert_subject_name: string
        :param hmac_key: If using HMAC, a string containing the shared secret.
        :type hmac_key: string
        :param validate_schema: Whether to validate **data** against the XML Signature schema.
        :type validate_schema: boolean
        :param parser: Custom XML parser instance to use when parsing **data**.
        :type parser: :py:class:`lxml.etree.XMLParser` compatible parser
        :param uri_resolver: Function to use to resolve reference URIs that don't start with "#".
        :type uri_resolver: callable
        :param id_attribute:
            Name of the attribute whose value ``URI`` refers to. By default, SignXML will search for "Id", then "ID".
        :type id_attribute: string
        :param expect_references:
            Number of references to expect in the signature. If this is not 1, an array of VerifyResults is returned.
            If set to a non-integer, any number of references is accepted (otherwise a mismatch raises an error).
        :type expect_references: int or boolean

        :raises: :py:class:`cryptography.exceptions.InvalidSignature`

        :returns: VerifyResult object with the signed data, signed xml and signature xml
        :rtype: VerifyResult

        """
        self.hmac_key = hmac_key
        self.require_x509 = require_x509
        self.x509_cert = x509_cert
        self._parser = parser

        if x509_cert:
            self.require_x509 = True

        if id_attribute is not None:
            self.id_attributes = (id_attribute, )

        root = self.get_root(data)
        if root.tag == ds_tag("Signature"):
            signature_ref = root
        else:
            signature_ref = self._find(root, "Signature", anywhere=True)

        # HACK: deep copy won't keep root's namespaces
        signature = fromstring(etree.tostring(signature_ref), parser=parser)

        if validate_schema:
            self.schema().assertValid(signature)

        signed_info = self._find(signature, "SignedInfo")
        c14n_method = self._find(signed_info, "CanonicalizationMethod")
        c14n_algorithm = c14n_method.get("Algorithm")
        signature_method = self._find(signed_info, "SignatureMethod")
        signature_value = self._find(signature, "SignatureValue")
        signature_alg = signature_method.get("Algorithm")
        raw_signature = b64decode(signature_value.text)
        x509_data = signature.find("ds:KeyInfo/ds:X509Data", namespaces=namespaces)
        signed_info_c14n = self._c14n(signed_info, algorithm=c14n_algorithm)

        if x509_data is not None or self.require_x509:
            from OpenSSL.crypto import load_certificate, X509, FILETYPE_PEM, verify, Error as OpenSSLCryptoError

            if self.x509_cert is None:
                if x509_data is None:
                    raise InvalidInput("Expected a X.509 certificate based signature")
                certs = [cert.text for cert in self._findall(x509_data, "X509Certificate")]
                if not certs:
                    msg = "Expected to find an X509Certificate element in the signature (X509SubjectName, X509SKI are not supported)"  # noqa
                    raise InvalidInput(msg)
                cert_chain = [load_certificate(FILETYPE_PEM, add_pem_header(cert)) for cert in certs]
                signing_cert = verify_x509_cert_chain(cert_chain, ca_pem_file=ca_pem_file, ca_path=ca_path)
            elif isinstance(self.x509_cert, X509):
                signing_cert = self.x509_cert
            else:
                signing_cert = load_certificate(FILETYPE_PEM, add_pem_header(self.x509_cert))

            if cert_subject_name and signing_cert.get_subject().commonName != cert_subject_name:
                raise InvalidSignature("Certificate subject common name mismatch")

            signature_digest_method = self._get_signature_digest_method(signature_alg).name
            try:
                verify(signing_cert, raw_signature, signed_info_c14n, signature_digest_method)
            except OpenSSLCryptoError as e:
                try:
                    lib, func, reason = e.message[0]
                except Exception:
                    reason = e
                raise InvalidSignature("Signature verification failed: {}".format(reason))
            # TODO: CN verification goes here
            # TODO: require one of the following to be set: either x509_cert or (ca_pem_file or ca_path) or common_name
            # Use ssl.match_hostname or code from it to perform match
        elif "hmac-sha" in signature_alg:
            if self.hmac_key is None:
                raise InvalidInput('Parameter "hmac_key" is required when verifying a HMAC signature')

            from cryptography.hazmat.primitives.hmac import HMAC
            signer = HMAC(key=ensure_bytes(self.hmac_key),
                          algorithm=self._get_hmac_digest_method(signature_alg),
                          backend=default_backend())
            signer.update(signed_info_c14n)
            if raw_signature != signer.finalize():
                raise InvalidSignature("Signature mismatch (HMAC)")
        else:
            key_value = signature.find("ds:KeyInfo/ds:KeyValue", namespaces=namespaces)
            if key_value is None:
                raise InvalidInput("Expected to find either KeyValue or X509Data XML element in KeyInfo")

            self._verify_signature_with_pubkey(signed_info_c14n, raw_signature, key_value, signature_alg)

        verify_results = []
        for reference in self._findall(signed_info, "Reference"):
            transforms = self._find(reference, "Transforms", require=False)
            digest_algorithm = self._find(reference, "DigestMethod").get("Algorithm")
            digest_value = self._find(reference, "DigestValue")
            payload = self._resolve_reference(root, reference, uri_resolver=uri_resolver)
            payload_c14n = self._apply_transforms(payload, transforms, signature_ref, c14n_algorithm)
            if digest_value.text != self._get_digest(payload_c14n, self._get_digest_method(digest_algorithm)):
                raise InvalidDigest("Digest mismatch for reference {}".format(len(verify_results)))

            # We return the signed XML (and only that) to ensure no access to unsigned data happens
            try:
                payload_c14n_xml = fromstring(payload_c14n)
            except etree.XMLSyntaxError:
                payload_c14n_xml = None
            verify_results.append(VerifyResult(payload_c14n, payload_c14n_xml, signature))

        if type(expect_references) is int and len(verify_results) != expect_references:
            msg = "Expected to find {} references, but found {}"
            raise InvalidSignature(msg.format(expect_references, len(verify_results)))

        return verify_results if expect_references > 1 else verify_results[0]

    def _get_long(self, element, query, require=True):
        result = self._find(element, query, require=require)
        if result is not None:
            result = bytes_to_long(b64decode(result.text))
        return result



import sys
import decimal
from decimal import Context

PY3 = sys.version_info[0] == 3
PY2 = sys.version_info[0] == 2


if PY3:
    long = int
    xrange = range
else:
    long = long  # pylint: disable=long-builtin
    xrange = xrange  # pylint: disable=xrange-builtin

# unicode / binary types
if PY3:
    text_type = str
    binary_type = bytes
    string_types = (str,)
    unichr = chr
    def maybe_decode(x):
        return x.decode()
    def maybe_encode(x):
        return x.encode()
    def maybe_chr(x):
        return x
    def maybe_ord(x):
        return x
else:
    text_type = unicode  # pylint: disable=unicode-builtin, undefined-variable
    binary_type = str
    string_types = (
        basestring,  # pylint: disable=basestring-builtin, undefined-variable
    )
    unichr = unichr  # pylint: disable=unichr-builtin
    def maybe_decode(x):
        return x
    def maybe_encode(x):
        return x
    def maybe_chr(x):
        return chr(x)
    def maybe_ord(x):
        return ord(x)


def round_py2_compat(what):
    """
    Python 2 and Python 3 use different rounding strategies in round(). This
    function ensures that results are python2/3 compatible and backward
    compatible with previous py2 releases
    :param what: float
    :return: rounded long
    """
    d = Context(
        prec=len(str(long(what))),  # round to integer with max precision
        rounding=decimal.ROUND_HALF_UP
    ).create_decimal(str(what))  # str(): python 2.6 compat
    return long(d)



from modelcommon.sitef import PaymentCodes
from pos_model import TenderType

from model import TPag


class SitefTPag(object):

    def __init__(
        self,   # type: TPag
    ):

        self.t_pag_dict = {
            "00001": TPag.VALE_ALIMENTACAO,
            "10001": TPag.VALE_ALIMENTACAO,
            "10002": TPag.VALE_ALIMENTACAO,
            "10003": TPag.VALE_ALIMENTACAO,
            "10005": TPag.VALE_ALIMENTACAO,
            "10006": TPag.VALE_ALIMENTACAO,
            "10007": TPag.VALE_ALIMENTACAO,
            "10008": TPag.VALE_ALIMENTACAO,
            "10009": TPag.VALE_ALIMENTACAO,
            "10010": TPag.VALE_ALIMENTACAO,
            "10011": TPag.VALE_ALIMENTACAO,
            "10021": TPag.VALE_REFEICAO,
            "10022": TPag.VALE_ALIMENTACAO,
            "10023": TPag.VALE_ALIMENTACAO,
            "10024": TPag.VALE_REFEICAO,
            "10026": TPag.VALE_ALIMENTACAO,
            "10027": TPag.VALE_ALIMENTACAO,
            "10028": TPag.VALE_REFEICAO,
            "10029": TPag.VALE_ALIMENTACAO,
            "10030": TPag.VALE_PRESENTE,
            "10031": TPag.VALE_PRESENTE,
            "10032": TPag.VALE_ALIMENTACAO,
            "10033": TPag.VALE_COMBUSTIVEL,
            "10051": TPag.VALE_ALIMENTACAO,
            "10053": TPag.VALE_ALIMENTACAO,
            "10054": TPag.VALE_COMBUSTIVEL,
            "10055": TPag.VALE_ALIMENTACAO,
            "60110024": TPag.PIX,
        }

    def get_t_pag(
        self,               # type: SitefTPag
        card_type,          # type: str
        payment_code,       # type: int
        electronic_type,    # type: int
    ):
        # type: (...) -> str

        t_pag = self._get_t_pag_from_payment(
            payment_code=payment_code,
            electronic_type=electronic_type,
        )

        if t_pag == TPag.CARTEIRA_DIGITAL:
            return self.t_pag_dict.get(card_type, TPag.CARTEIRA_DIGITAL)
        if t_pag == TPag.DEBITO:
            return self.t_pag_dict.get(card_type, TPag.DEBITO)
        return t_pag

    @staticmethod
    def _get_t_pag_from_payment(
        payment_code,       # type: int
        electronic_type,    # type: int
    ):
        # type: (...) -> str

        if payment_code == PaymentCodes.debit:
            return TPag.DEBITO
        if payment_code == PaymentCodes.credit:
            return TPag.VALE_PRESENTE if electronic_type == TenderType.gift_card else TPag.CREDITO
        if payment_code == PaymentCodes.digital_wallet:
            return TPag.CARTEIRA_DIGITAL

        return ""



from typing import (
    Dict,
    Optional,
    Union,
)


def update_tender_details_with_discount_data(
    discount_data,      # type: Optional[Dict[str, Union[str, float]]]
    tender_details,     # type: Dict[str, Union[str, float, int, Dict]]
):
    # type: (...) -> None

    if not discount_data:
        return

    benefit_id = discount_data.get("benefit_id")
    if benefit_id:
        tender_details["benefitId"] = benefit_id

    loyalty_id = discount_data.get("loyalty_id")
    if loyalty_id is not None:
        tender_details["loyaltyId"] = loyalty_id



from ._OrderChangerProductionBox import OrderChangerProductionBox
from ..model import Item


class NotShowOnKitchenAddTagBox(OrderChangerProductionBox):

    def __init__(
        self,
        name,
        sons,
        tag_to_add,
        product_repository,
        logger=None,
    ):
        super(NotShowOnKitchenAddTagBox, self).__init__(
            name=name,
            sons=sons,
            logger=logger,
        )

        self.forbidden_part_codes = product_repository.get_not_show_on_kitchen()
        self.tag_to_add = tag_to_add

    def change_order(self, order):
        for item in order.items:
            self.handle_item(item)

        return order

    def handle_item(
        self,       # type: NotShowOnKitchenAddTagBox
        item,       # type: Item
    ):
        # type: (...) -> None

        if item.part_code in self.forbidden_part_codes:
            item.add_tag(self.tag_to_add)

        for son in item.items:
            self.handle_item(son)



from xml.etree import cElementTree as eTree

from modelcommon.fiscalwrapper import FiscalizationException
from mw_helper import show_message_dialog
from typing import Optional

from sysactions import action, get_model, get_current_order

from custom_sysactions import user_must_be_logged
from . import get_error_fiscal_message
from .. import logger
from application.interactor import FiscalWrapperInteractor


@action
@user_must_be_logged
def check_can_fiscalize_order_action(
    pos_id,         # type: str
    order=None,     # type: Optional[eTree.XML]
):
    # type: (...) -> bool

    model = get_model(posid=pos_id)
    if order is None:
        order = get_current_order(model=model)

    try:
        FiscalWrapperInteractor().can_fiscalize_order(order=order)
        return True
    except FiscalizationException as ex:
        logger.exception(msg="Error check can fiscalize order", exc_info=ex)
        _show_message_error(
            ex=ex,
            model=model,
            pos_id=pos_id,
        )
        return False


def _show_message_error(
    ex,         # type: Exception
    model,      # type: Optional[eTree.XML]
    pos_id,     # type: str
):
    # type: (...) -> None

    message = get_error_fiscal_message(
        model=model,
        exception=ex,
    )
    show_message_dialog(
        pos_id=pos_id,
        title="$ERROR",
        message=message,
    )



import operator

from jmespath import functions
from jmespath.compat import string_type
from numbers import Number


def _equals(x, y):
    if _is_special_integer_case(x, y):
        return False
    else:
        return x == y


def _is_special_integer_case(x, y):
    # We need to special case comparing 0 or 1 to
    # True/False.  While normally comparing any
    # integer other than 0/1 to True/False will always
    # return False.  However 0/1 have this:
    # >>> 0 == True
    # False
    # >>> 0 == False
    # True
    # >>> 1 == True
    # True
    # >>> 1 == False
    # False
    #
    # Also need to consider that:
    # >>> 0 in [True, False]
    # True
    if type(x) is int and (x == 0 or x == 1):
        return y is True or y is False
    elif type(y) is int and (y == 0 or y == 1):
        return x is True or x is False


def _is_comparable(x):
    # The spec doesn't officially support string types yet,
    # but enough people are relying on this behavior that
    # it's been added back.  This should eventually become
    # part of the official spec.
    return _is_actual_number(x) or isinstance(x, string_type)


def _is_actual_number(x):
    # We need to handle python's quirkiness with booleans,
    # specifically:
    #
    # >>> isinstance(False, int)
    # True
    # >>> isinstance(True, int)
    # True
    if x is True or x is False:
        return False
    return isinstance(x, Number)


class Options(object):
    """Options to control how a JMESPath function is evaluated."""
    def __init__(self, dict_cls=None, custom_functions=None):
        #: The class to use when creating a dict.  The interpreter
        #  may create dictionaries during the evaluation of a JMESPath
        #  expression.  For example, a multi-select hash will
        #  create a dictionary.  By default we use a dict() type.
        #  You can set this value to change what dict type is used.
        #  The most common reason you would change this is if you
        #  want to set a collections.OrderedDict so that you can
        #  have predictable key ordering.
        self.dict_cls = dict_cls
        self.custom_functions = custom_functions


class _Expression(object):
    def __init__(self, expression, interpreter):
        self.expression = expression
        self.interpreter = interpreter

    def visit(self, node, *args, **kwargs):
        return self.interpreter.visit(node, *args, **kwargs)


class Visitor(object):
    def __init__(self):
        self._method_cache = {}

    def visit(self, node, *args, **kwargs):
        node_type = node['type']
        method = self._method_cache.get(node_type)
        if method is None:
            method = getattr(
                self, 'visit_%s' % node['type'], self.default_visit)
            self._method_cache[node_type] = method
        return method(node, *args, **kwargs)

    def default_visit(self, node, *args, **kwargs):
        raise NotImplementedError("default_visit")


class TreeInterpreter(Visitor):
    COMPARATOR_FUNC = {
        'eq': _equals,
        'ne': lambda x, y: not _equals(x, y),
        'lt': operator.lt,
        'gt': operator.gt,
        'lte': operator.le,
        'gte': operator.ge
    }
    _EQUALITY_OPS = ['eq', 'ne']
    MAP_TYPE = dict

    def __init__(self, options=None):
        super(TreeInterpreter, self).__init__()
        self._dict_cls = self.MAP_TYPE
        if options is None:
            options = Options()
        self._options = options
        if options.dict_cls is not None:
            self._dict_cls = self._options.dict_cls
        if options.custom_functions is not None:
            self._functions = self._options.custom_functions
        else:
            self._functions = functions.Functions()

    def default_visit(self, node, *args, **kwargs):
        raise NotImplementedError(node['type'])

    def visit_subexpression(self, node, value):
        result = value
        for node in node['children']:
            result = self.visit(node, result)
        return result

    def visit_field(self, node, value):
        try:
            return value.get(node['value'])
        except AttributeError:
            return None

    def visit_comparator(self, node, value):
        # Common case: comparator is == or !=
        comparator_func = self.COMPARATOR_FUNC[node['value']]
        if node['value'] in self._EQUALITY_OPS:
            return comparator_func(
                self.visit(node['children'][0], value),
                self.visit(node['children'][1], value)
            )
        else:
            # Ordering operators are only valid for numbers.
            # Evaluating any other type with a comparison operator
            # will yield a None value.
            left = self.visit(node['children'][0], value)
            right = self.visit(node['children'][1], value)
            num_types = (int, float)
            if not (_is_comparable(left) and
                    _is_comparable(right)):
                return None
            return comparator_func(left, right)

    def visit_current(self, node, value):
        return value

    def visit_expref(self, node, value):
        return _Expression(node['children'][0], self)

    def visit_function_expression(self, node, value):
        resolved_args = []
        for child in node['children']:
            current = self.visit(child, value)
            resolved_args.append(current)
        return self._functions.call_function(node['value'], resolved_args)

    def visit_filter_projection(self, node, value):
        base = self.visit(node['children'][0], value)
        if not isinstance(base, list):
            return None
        comparator_node = node['children'][2]
        collected = []
        for element in base:
            if self._is_true(self.visit(comparator_node, element)):
                current = self.visit(node['children'][1], element)
                if current is not None:
                    collected.append(current)
        return collected

    def visit_flatten(self, node, value):
        base = self.visit(node['children'][0], value)
        if not isinstance(base, list):
            # Can't flatten the object if it's not a list.
            return None
        merged_list = []
        for element in base:
            if isinstance(element, list):
                merged_list.extend(element)
            else:
                merged_list.append(element)
        return merged_list

    def visit_identity(self, node, value):
        return value

    def visit_index(self, node, value):
        # Even though we can index strings, we don't
        # want to support that.
        if not isinstance(value, list):
            return None
        try:
            return value[node['value']]
        except IndexError:
            return None

    def visit_index_expression(self, node, value):
        result = value
        for node in node['children']:
            result = self.visit(node, result)
        return result

    def visit_slice(self, node, value):
        if not isinstance(value, list):
            return None
        s = slice(*node['children'])
        return value[s]

    def visit_key_val_pair(self, node, value):
        return self.visit(node['children'][0], value)

    def visit_literal(self, node, value):
        return node['value']

    def visit_multi_select_dict(self, node, value):
        if value is None:
            return None
        collected = self._dict_cls()
        for child in node['children']:
            collected[child['value']] = self.visit(child, value)
        return collected

    def visit_multi_select_list(self, node, value):
        if value is None:
            return None
        collected = []
        for child in node['children']:
            collected.append(self.visit(child, value))
        return collected

    def visit_or_expression(self, node, value):
        matched = self.visit(node['children'][0], value)
        if self._is_false(matched):
            matched = self.visit(node['children'][1], value)
        return matched

    def visit_and_expression(self, node, value):
        matched = self.visit(node['children'][0], value)
        if self._is_false(matched):
            return matched
        return self.visit(node['children'][1], value)

    def visit_not_expression(self, node, value):
        original_result = self.visit(node['children'][0], value)
        if type(original_result) is int and original_result == 0:
            # Special case for 0, !0 should be false, not true.
            # 0 is not a special cased integer in jmespath.
            return False
        return not original_result

    def visit_pipe(self, node, value):
        result = value
        for node in node['children']:
            result = self.visit(node, result)
        return result

    def visit_projection(self, node, value):
        base = self.visit(node['children'][0], value)
        if not isinstance(base, list):
            return None
        collected = []
        for element in base:
            current = self.visit(node['children'][1], element)
            if current is not None:
                collected.append(current)
        return collected

    def visit_value_projection(self, node, value):
        base = self.visit(node['children'][0], value)
        try:
            base = base.values()
        except AttributeError:
            return None
        collected = []
        for element in base:
            current = self.visit(node['children'][1], element)
            if current is not None:
                collected.append(current)
        return collected

    def _is_false(self, value):
        # This looks weird, but we're explicitly using equality checks
        # because the truth/false values are different between
        # python and jmespath.
        return (value == '' or value == [] or value == {} or value is None or
                value is False)

    def _is_true(self, value):
        return not self._is_false(value)


class GraphvizVisitor(Visitor):
    def __init__(self):
        super(GraphvizVisitor, self).__init__()
        self._lines = []
        self._count = 1

    def visit(self, node, *args, **kwargs):
        self._lines.append('digraph AST {')
        current = '%s%s' % (node['type'], self._count)
        self._count += 1
        self._visit(node, current)
        self._lines.append('}')
        return '\n'.join(self._lines)

    def _visit(self, node, current):
        self._lines.append('%s [label="%s(%s)"]' % (
            current, node['type'], node.get('value', '')))
        for child in node.get('children', []):
            child_name = '%s%s' % (child['type'], self._count)
            self._count += 1
            self._lines.append('  %s -> %s' % (current, child_name))
            self._visit(child, child_name)
